import numpy as np
from operator import itemgetter
from itertools import groupby
from numpy.core.fromnumeric import mean
from scipy.optimize import minimize
from ligeor.utils.lcutils import *
from ligeor.models import Model


class Polyfit(Model):

    def __init__(self, filename='', phases=[], fluxes=[], sigmas=[], n_downsample=1, usecols=[0,1,2], delimiter=',',
                        phase_folded=True, period=1, t0=0,
                        xmin=-0.5, xmax=0.5, polyorder=2, chain_length=4):
        '''
        Initializes a polyfit with data and fit parameters.

        Parameters
        ----------
        filename: str
            Filename of light curve to load (if phases, fluxes and sigmas not provided)
        phases: array-like
            Input orbital phases, must be on the range [-0.5,0.5]
        fluxes: array-like
            Input fluxes
        sigmas: array-like
            Input sigmas (corresponding flux uncertainities)
        n_downsample: int
            Number of data points to skip if downsampling
        xmin: float
            Minimum orbital phase of the polyfit
        xmax: float
            Maximum orbital phase of the polyfit
        polyorder: int
            Polynomial order of each chain
        chain_length: int
            Length of the chain
        '''

        if polyorder != 2:
            raise NotImplemented('polyorder != 2 is not yet implemented.')
        if chain_length != 4:
            raise NotImplemented('chain_length != 4 is not yet implemented.')


        super(Polyfit, self).__init__(phases, fluxes, sigmas, filename, n_downsample, usecols, delimiter, phase_folded, period, t0)


        # for now, build a data attribute to match the Polychain methods
        self.data = np.array([self.phases, self.fluxes, self.sigmas]).T

        self.xmin = xmin
        self.xmax = xmax
        self.polyorder = polyorder
        self.chain_length = chain_length
        # self.sdata = self.data.copy()#[self.data[:,0].argsort()]
        self.sdata = self.data[self.data[:,0].argsort()]

        
    def _find_knots(self, min_chain_length=8, verbose=False):
        mean = self.sdata[:,1].mean()
        lt = np.where(self.sdata[:,1]<mean)[0]
        
        chains = []
        for k, g in groupby(enumerate(lt), lambda ix: ix[0]-ix[1]):
            chain = list(map(itemgetter(1), g))
            if len(chain) > min_chain_length:
                chains.append(chain)
        if verbose:
            print(f'{len(chains)} chains found.')
        if len(chains) < 2:
            self.knots = np.array((-0.4, -0.1, 0.1, 0.4))
        elif len(chains) == 2:
            self.knots = np.sort((self.sdata[chains[0][0],0], self.sdata[chains[0][-1],0], self.sdata[chains[1][0],0], self.sdata[chains[1][-1],0]))
        else:
            lengths = np.array([len(chain) for chain in chains])
            l = np.argsort(lengths)[::-1]
            self.knots = np.sort((self.sdata[chains[l[0]][0],0], self.sdata[chains[l[0]][-1],0], self.sdata[chains[l[1]][0],0], self.sdata[chains[l[1]][-1],0]))

        return self.knots
    
    def _find_segments(self, knots):
        self.sdata = self.data.copy()
        self.sdata[:,0][self.sdata[:,0] < knots[0]] += 1
        self.sdata = self.sdata[self.sdata[:,0].argsort()]

        segs = [np.argmax(self.sdata[:,0]>knot) for knot in knots[1:]] + [len(self.sdata)]
        return segs

    def _build_A_matrix(self, knots, segs):
        self.A = np.zeros(shape=(len(self.sdata), 8))

        # segment 1:
        self.A[:segs[0],0] = self.sdata[:segs[0],0]**2
        self.A[:segs[0],1] = self.sdata[:segs[0],0]
        self.A[:segs[0],2] = 1.0

        # segment 2:
        self.A[segs[0]:segs[1],0] = knots[1]**2
        self.A[segs[0]:segs[1],1] = knots[1]
        self.A[segs[0]:segs[1],2] = 1.0
        self.A[segs[0]:segs[1],3] = self.sdata[segs[0]:segs[1],0]**2-knots[1]**2
        self.A[segs[0]:segs[1],4] = self.sdata[segs[0]:segs[1],0]-knots[1]

        # segment 3:
        self.A[segs[1]:segs[2],0] = knots[1]**2
        self.A[segs[1]:segs[2],1] = knots[1]
        self.A[segs[1]:segs[2],2] = 1.0
        self.A[segs[1]:segs[2],3] = knots[2]**2-knots[1]**2
        self.A[segs[1]:segs[2],4] = knots[2]-knots[1]
        self.A[segs[1]:segs[2],5] = self.sdata[segs[1]:segs[2],0]**2-knots[2]**2
        self.A[segs[1]:segs[2],6] = self.sdata[segs[1]:segs[2],0]-knots[2]

        # segment 4:
        x = (self.sdata[segs[2]:segs[3],0]-knots[3])/(knots[3]-knots[0]-1)
        self.A[segs[2]:segs[3],0] = knots[1]**2 + x*(knots[1]**2-knots[0]**2)
        self.A[segs[2]:segs[3],1] = knots[1] + x*(knots[1]-knots[0])
        self.A[segs[2]:segs[3],2] = 1.0
        self.A[segs[2]:segs[3],3] = knots[2]**2-knots[1]**2 + x*(knots[2]**2-knots[1]**2)
        self.A[segs[2]:segs[3],4] = knots[2]-knots[1] + x*(knots[2]-knots[1])
        self.A[segs[2]:segs[3],5] = knots[3]**2-knots[2]**2 + x*(knots[3]**2-knots[2]**2)
        self.A[segs[2]:segs[3],6] = knots[3]-knots[2] + x*(knots[3]-knots[2])
        self.A[segs[2]:segs[3],7] = self.sdata[segs[2]:segs[3],0]**2-knots[3]**2 + x*((knots[0]+1)**2-knots[3]**2)

    def _fit_chain(self, knots, min_pts_per_segment=5, return_ck=False):
        # we need to sort the knots first because the minimizer can criss-cross them:
        knots = np.sort(knots)
        
        segs = self._find_segments(knots)
        if np.ediff1d(segs).min() < min_pts_per_segment or np.any(knots < self.xmin):
            if return_ck:
                return None, 1e10
            return 1e10

        self._build_A_matrix(knots, segs)
        ck, ssr, rank, svd = np.linalg.lstsq(self.A, self.sdata[:,1], rcond=None)

        # safety switch when things go wrong:
        if len(ssr) == 0:
            ssr = 1e10

        if return_ck:
            return ck, ssr
        # print(ck, ssr, rank, svd, knots, segs)
        return ssr
    
    def _chain_coeffs(self, ck, verbose=False):
        if ck is None:
            self.coeffs = None
            return
        
        c0 = ck[2]
        c1 = (ck[0]-ck[3])*self.knots[1]**2 + (ck[1]-ck[4])*self.knots[1] + c0
        c2 = (ck[3]-ck[5])*self.knots[2]**2 + (ck[4]-ck[6])*self.knots[2] + c1
        b3 = 1./(self.knots[3]-self.knots[0]-1)*(ck[0]*(self.knots[1]**2-self.knots[0]**2)+ck[3]*(self.knots[2]**2-self.knots[1]**2)+ck[5]*(self.knots[3]**2-self.knots[2]**2)+ck[7]*((self.knots[0]+1)**2-self.knots[3]**2)+ck[1]*(self.knots[1]-self.knots[0])+ck[4]*(self.knots[2]-self.knots[1])+ck[6]*(self.knots[3]-self.knots[2]))
        c3 = (ck[5]-ck[7])*self.knots[3]**2 + (ck[6]-b3)*self.knots[3] + c2

        if verbose:
            print(f'segment 1:\n {self.knots[0]} < x <= {self.knots[1]}, a0={ck[0]} b0={ck[1]} c0={c0}')
            print(f'segment 2:\n {self.knots[1]} < x <= {self.knots[2]}, a1={ck[3]} b1={ck[4]} c1={c1}')
            print(f'segment 3:\n {self.knots[2]} < x <= {self.knots[3]}, a2={ck[5]} b2={ck[6]} c2={c2}')
            print(f'segment 4:\n {self.knots[3]} < x <= {self.knots[0]+1}, a3={ck[7]} b3={b3} c3={c3}')

        self.coeffs = ((ck[0], ck[1], c0), (ck[3], ck[4], c1), (ck[5], ck[6], c2), (ck[7], b3, c3))
    
    def _chain_extremes(self):
        if self.coeffs is None:
            self.extremes = None
            return

        exts = [-c[1]/2/c[0] for c in self.coeffs]
        knots = np.concatenate((self.knots, [self.knots[0]+1]))
        for k in range(4):
            if exts[k] < knots[k] or exts[k] > knots[k+1]:
                exts[k] = np.nan
        self.extremes = np.array(exts)
    
    def _remap_1d(self, d, sort=True):
        while len(d[d<self.xmin]) > 0:
            d[d<self.xmin] += 1
        while len(d[d>self.xmax]) > 0:
            d[d>self.xmax] -= 1

        if sort:
            d.sort()

        return d

    def _remap(self, d, sort=True):
        ncols = 1 if len(d.shape) == 1 else d.shape[1]
        x = d if ncols == 1 else d[:,0]

        while len(x[x<self.xmin]) > 0:
            x[x<self.xmin] += 1
        while len(x[x>self.xmax]) > 0:
            x[x>self.xmax] -= 1

        if sort and ncols > 1:
            d = d[d[:,0].argsort()]
        else:
            d.sort()

        return d      
    
    def fit(self, min_chain_length=8, min_pts_per_segment=10, method='Nelder-Mead', knots=None, coeffs = None, verbose=False):
        if knots is None:
            self._find_knots(min_chain_length=min_chain_length, verbose=verbose)
        else:
            self.knots = knots
        
        if coeffs is None:
            solution = minimize(self._fit_chain, self.knots, args=(min_pts_per_segment,), method=method)
            self.knots = solution.x
            ck, self.ssr = self._fit_chain(self.knots, min_pts_per_segment=min_pts_per_segment, return_ck=True)
            self._chain_coeffs(ck)
        else:
            self.coeffs = coeffs

        self._chain_extremes()
        self.model = self.fv(x=self.phases)
        
        return self.knots, self.coeffs, self.extremes
    
    def fv(self, x):
        if self.coeffs is None:
            return None
        x[x<self.knots[0]] += 1
        y = np.empty_like(x)
        for k in range(len(self.knots)-1):
            s = (x>=self.knots[k]) & (x<self.knots[k+1])
            y[s] = self.coeffs[k][0]*x[s]**2 + self.coeffs[k][1]*x[s] + self.coeffs[k][2]
        s = x>=self.knots[3]
        y[s] = self.coeffs[3][0]*x[s]**2 + self.coeffs[3][1]*x[s] + self.coeffs[3][2]

        return y


    def plot(self, x, savefig=None, show=True):
        import matplotlib.pyplot as plt

        if self.coeffs is None:
            return None

        if x is None:
            x = self.sdata[:,0]

        knot_fvs = self.fv(self.knots.copy())
        knots = self._remap_1d(self.knots.copy(), sort=False)

        y = self.fv(x)
        d = np.vstack((x, y)).T

        self._remap(self.sdata)
        d = self._remap(d)
        self._remap_1d(self.extremes)

        # print(knots)
        # print(knot_fvs)

        plt.plot(self.phases, self.fluxes, 'b.')
        plt.plot(knots, knot_fvs, 'ms')
        plt.plot(d[:,0], d[:,1], 'r-')

        for k in range(4):
            plt.axvline(self.extremes[k], ls='--')
        
        if savefig is not None:
            plt.savefig(savefig)

        if show:
            plt.show()
        else:
            plt.clf()


    def compute_eclipse_params(self, interactive=False):
        '''
        Compute the positions, widths and depths of the eclipses 
        based on the polyfit solution.

        The eclipse parameters are computed as following:
        - eclipses are first identified as the chains corresponding to the
          two deepest minima
        - eclipse positions are set to the minima positions
        - eclipse widths are the difference between the chain knots 
        - eclipse depths are the difference between the mean function values at the
          knots and function values at the minima.

        Returns
        -------
        results: dict
            A dictionary of the eclipse paramter values.
        '''
        ext_x = self.extremes
        ext_y = self.fv(x=ext_x)
        knots_x = self.knots
        knots_extended = np.hstack((self.knots-1, self.knots, self.knots+1))

        xs_all = np.hstack((np.array(ext_x), np.array(knots_x))).flatten()
        ys_all = self.fv(x=xs_all)
        ys_labels = np.array(['ext', 'ext', 'ext', 'ext', 'knot', 'knot', 'knot', 'knot'])
        sort = np.argsort(ys_all)
        if ys_labels[sort][0] == 'knot' and ys_labels[sort][1] == 'ext':
            # no primary eclipse, just log details of secondary
            primary_pos = xs_all[sort][0]
            eclipse_arg = np.argwhere(ext_x == xs_all[sort][1]).flatten()[0]
            knots_ecl = np.array([knots_extended[eclipse_arg+4], knots_extended[eclipse_arg+5]])
            mean_outofecl = self.fv(knots_ecl).mean()
            
            self.eclipse_params = {
            'primary_width': np.nan,
            'secondary_width': np.abs(knots_ecl[1]-knots_ecl[0]),
            'primary_position': np.nan,
            'secondary_position': ext_x[eclipse_arg],
            'primary_depth': np.nan,
            'secondary_depth': mean_outofecl - ext_y[eclipse_arg],
            'eclipse_edges': np.hstack((np.array([np.nan,np.nan]), knots_ecl)),
            'eclipse_coeffs': [np.nan*np.ones((3)), self.coeffs[eclipse_arg]]
        }
            
        if ys_labels[sort][0] == 'ext' and ys_labels[sort][1] == 'knot':
                # no primary eclipse, just log details of secondary
            secondary_pos = xs_all[sort][1]
            eclipse_arg = np.argwhere(ext_x == xs_all[sort][0]).flatten()[0]
            knots_ecl = np.array([knots_extended[eclipse_arg+4], knots_extended[eclipse_arg+5]])
            mean_outofecl = self.fv(knots_ecl).mean()
            
            self.eclipse_params = {
            'primary_width': np.abs(knots_ecl[1]-knots_ecl[0]),
            'secondary_width': np.nan,
            'primary_position': ext_x[eclipse_arg],
            'secondary_position': np.nan,
            'primary_depth': mean_outofecl - ext_y[eclipse_arg],
            'secondary_depth': np.nan,
            'eclipse_edges': np.hstack((knots_ecl, np.array([np.nan,np.nan]))),
            'eclipse_coeffs': [self.coeffs[eclipse_arg], np.nan*np.ones((3))]
        }
        
        
        if ys_labels[sort][0] == 'ext' and ys_labels[sort][1] == 'ext':
            # fv for x=np.nan doesn't return nan!
            #TODO: troubleshoot .fv for a discrete set of values and why np.nan results in a value
            ext_y[np.isnan(self.extremes)] = np.nan
            eclipse_args = np.argsort(ext_y)[:2]

            # let's extend the knots array left and right so it's continuous (for computing the width)
            knots1 = np.array([knots_extended[eclipse_args[0]+4], knots_extended[eclipse_args[0]+5]])
            knots2 = np.array([knots_extended[eclipse_args[1]+4], knots_extended[eclipse_args[1]+5]])

            mean_outofecl = self.fv(np.hstack((knots1, knots2))).mean()

            self.eclipse_params = {
                'primary_width': np.abs(knots1[1]-knots1[0]),
                'secondary_width': np.abs(knots2[1]-knots2[0]),
                'primary_position': ext_x[eclipse_args[0]],
                'secondary_position': ext_x[eclipse_args[1]],
                'primary_depth': mean_outofecl - ext_y[eclipse_args[0]],
                'secondary_depth': mean_outofecl - ext_y[eclipse_args[1]],
                'eclipse_edges': np.hstack((knots1, knots2)),
                'eclipse_coeffs': [self.coeffs[eclipse_args[0]], self.coeffs[eclipse_args[1]]]
            }
        
        if ys_labels[sort][0] == 'knot' and ys_labels[sort][1] == 'knot':
                # no primary eclipse, just log details of secondary
            primary_pos = xs_all[sort][0]
            secondary_pos = xs_all[sort][1]

            self.eclipse_params = {
            'primary_width': np.nan,
            'secondary_width': np.nan,
            'primary_position': primary_pos,
            'secondary_position': secondary_pos,
            'primary_depth': np.nan,
            'secondary_depth': np.nan,
            'eclipse_edges': np.array([np.nan, np.nan, np.nan, np.nan]),
            'eclipse_coeffs': [np.nan*np.ones((3)), np.nan*np.ones((3))]
        }


        if interactive:
           self.interactive_eclipse()


        self.compute_eclipse_area(ecl=1)
        self.compute_eclipse_area(ecl=2)

        # self.eclipse_params = self.check_eclipses_credibility()

        # check if eclipses need to be swapped:
        if ~np.isnan(self.eclipse_params['secondary_depth']) and ~np.isnan(self.eclipse_params['primary_depth']): 

            if self.eclipse_params['secondary_depth'] > self.eclipse_params['primary_depth']:
                # the secondary is deeper than the primary, so we swap them
                pos1, d1, w1, edge1 = self.eclipse_params['secondary_position'], self.eclipse_params['secondary_depth'], self.eclipse_params['secondary_width'], self.eclipse_params['eclipse_edges'][2:]
                pos2, d2, w2, edge2 = self.eclipse_params['primary_position'], self.eclipse_params['primary_depth'], self.eclipse_params['primary_width'], self.eclipse_params['eclipse_edges'][:2]

                self.eclipse_params['primary_position'] = pos1 
                self.eclipse_params['primary_width'] = w1
                self.eclipse_params['primary_depth'] = d1
                self.eclipse_params['secondary_position'] = pos2 
                self.eclipse_params['secondary_width'] = w2
                self.eclipse_params['secondary_depth'] = d2
                self.eclipse_params['eclipse_edges'] = [edge1[0],edge1[1], edge2[0], edge2[1]]

        elif ~np.isnan(self.eclipse_params['secondary_depth']) and np.isnan(self.eclipse_params['primary_depth']):
            # there is only one eclipse and it's fitted by the "secondary", so we need to move it to primary
            pos1, d1, w1, edge1 = self.eclipse_params['secondary_position'], self.eclipse_params['secondary_depth'], self.eclipse_params['secondary_width'], self.eclipse_params['eclipse_edges'][2:]

            self.eclipse_params['primary_position'] = pos1 
            self.eclipse_params['primary_width'] = w1
            self.eclipse_params['primary_depth'] = d1
            self.eclipse_params['secondary_position'] = np.nan
            self.eclipse_params['secondary_width'] = np.nan
            self.eclipse_params['secondary_depth'] = np.nan
            self.eclipse_params['eclipse_edges'] = [edge1[0],edge1[1], np.nan, np.nan]

        return self.eclipse_params
        

    def compute_eclipse_area(self, ecl=1):
        '''
        Computes the area under an eclipse.

        An eclipse is defined as being positioned between the knots around the deepest minima.

        Parameters
        ----------
        ecl: int
            The eclipse whose area is to be computed (1 or 2)
        
        Returns
        -------
        eclipse_area: float
            The computed area under the chosen eclipse.
        '''

        if hasattr(self, 'eclipse_area'):
            pass
        else:
            self.eclipse_area = {}

        coeffs = self.eclipse_params['eclipse_coeffs'][ecl-1]
        if ecl == 1:
            edges = [self.eclipse_params['eclipse_edges'][0], self.eclipse_params['eclipse_edges'][1]]
        else:
            edges = [self.eclipse_params['eclipse_edges'][2], self.eclipse_params['eclipse_edges'][3]]

        self.eclipse_area[ecl] = coeffs[0]/3*(edges[1]**3-edges[0]**3) + coeffs[1]/2*(edges[1]**2-edges[0]**2) + coeffs[2]*(edges[1]-edges[0])



import numpy as np
import emcee 
from ligeor.utils.lcutils import *
from multiprocessing import Pool

class EmceeSampler(object):

    def __init__(self, filename, times, fluxes, sigmas,
                period_init, t0_init, n_downsample, initial_fit_kwargs = {}, **kwargs):
        '''
        Initializes a sampler for the light curve stored in 'filename'
        with determined initial values for the period and t0.

        Parameters
        ----------
        filename: str
            Filename to load the raw light curve from (expected format: times, fluxes, sigmas)
        times: array-like
            Time points of the observations
        fluxes: array-like
            Observed fluxes
        sigmas: array-like
            Uncertainities of the observed fluxes
        period_init: float
            Initial value of the period (from code/triage)
        t0_init: float
            Initial value of the time of superior conjunction (t0)
        n_downsample: int
            Number of data points to skip in raw light curve for downsampling
        nbins: int
            Number of bins (applies to the computed final model).
        
        Keyword arguments are passed on to load_lc.
            
        
        Attributes
        ----------
        _period_init: float
            Initial value of the orbital period
        _t0_init: float
            Initial value of the time of superior conjunction
        '''
        print(times)
        if len(times) == 0:
            if len(filename) == 0:
                raise ValueError("Must provide either a filename or data through times, fluxes and sigmas")
            else:
                self.__filename = filename
                self._period_init = period_init 
                self._t0_init = t0_init
                lc = load_lc(self.__filename, n_downsample=n_downsample, **kwargs)
                self._times, self._fluxes, self._sigmas = lc['times'], lc['fluxes'], lc['sigmas']
        else:
            # check length of phases and fluxes
            if len(times) != len(fluxes):
                raise ValueError('Length of array mismatch between times ({}) and fluxes ({})'.format(len(times), len(fluxes)))
            else:
                # sort just in case
                self._times, self._fluxes, self._sigmas = times[::n_downsample], fluxes[::n_downsample], sigmas[::n_downsample]

        self.initial_fit(**initial_fit_kwargs)
        
    def initial_fit(self, **kwargs):
        '''
        Runs an initial fit to the data with the chosen model (two-Gaussian or polyfit).
        
        Attributes
        ----------
        _func: str
            Name of the best fit model function.
        _model_params: array-like
            Labels of all model parameters
        _initial_fit: array-like
            Initial values of the model parameters from optimization.
        '''
        # overriden by subclass
        return None
    
    def logprob(self, values):
        '''
        Computes the logprobability of the sample.

        Parameters
        ----------
        values: array-like
            period (for phase folding) + model values
            
        '''
        # overriden by subclass
        return None

    def run_sampler(self, nwalkers=32, niters=2000, progress=True, emcee_kwargs = {}, sample_kwargs={}):
        '''
        Initializes and runs an emcee sampler.
        
        Parameters
        ----------
        nwalkers: int
            Number of walkers for emcee.
        niters: int
            Number of iterations to run.
        progress: bool
            If True, will output the progress (requires tqdm).
            
        Attributes
        ----------
        _sampler: object
            emcee.EnsembleSampler object, containing the MCMC samples, blobs and logprobabilities
        '''
        init_vals = self._initial_fit
        bestfit_vals = np.array([self._period_init, *init_vals])
        pos = bestfit_vals + 1e-4 * np.random.randn(nwalkers, len(bestfit_vals))
        nwalkers, ndim = pos.shape
        
        with Pool(1) as pool:
            self._sampler = emcee.EnsembleSampler(nwalkers, ndim, self.logprob, pool=pool, **emcee_kwargs)
            self._sampler.run_mcmc(pos, niters, progress=progress, **sample_kwargs)
    

    def compute_results(self, burnin = 1000, save_lc=True, save_file='', show=False, failed=False):
        '''
        Computes a summary of the results from the sampler.

        The results computed include: ephemerides, model parameters and eclipse parameters. 

        Parameters
        ----------
        sampler: object
            The emcee sampler, initialized and run with .run_sampler()
        burnin: int
            Number of initial iterations to discard.
        save_lc: bool
            If True, will save the light curve to a file.
        save_file: str
            Filename to save to, if save_lc = True.
        show: bool
            If True, will show plot of the resulting light curve.
        failed: bool
            If True, all computed values are np.nan
            
        Attributes
        ----------
        _period_mcmc: dict
            Period mean and sigmas, computed from the posterior distribution
        _t0_mcmc: dict
            t0 mean and sigmas, computed from the posterior distribution
        _eclipse_params: dict
            Means of the eclipse parameters, computed from the posterior distribution
        _eclipse_params_errs: dict
            Uncertainities of the eclipse parameters from the posterior distribution
        _model_values: dict
            Model parameter values from the posterior distribution
        _model_values_errs: dict
            Uncertainities of the model parameter values from the posterior distribution
        _chi2: float
            Mean chi2 of the model, given the data.
        '''

        if save_lc and len(save_file) == 0:
            raise ValueError('Please provide a file name to save the model to or set save_lc=False.')
        
        # store the rest of the model parameters and their uncertainties
        means, sigmas_low, sigmas_high, means_blobs, sigmas_low_blobs, sigmas_high_blobs = self.compute_ephemerides(self._sampler, burnin, failed=failed)
        self.compute_model(means, sigmas_low, sigmas_high, save_lc = save_lc, save_file=save_file, show=show, failed=failed)
        self.compute_eclipse_params(means_blobs, sigmas_low_blobs, sigmas_high_blobs, failed=failed)


    def compute_ephemerides(self, sampler, burnin, failed=False):
        '''
        Computes mean and standard deviation for the period and t0 from the sample.

        Parameters
        ----------
        sampler: object
            The emcee sampler, initialized and run with .run_sampler()
        burnin: int
            Number of initial iterations to discard.
        '''

        if failed:
            self._period_mcmc = {'value': np.nan,
                    'sigma_low': np.nan, 
                    'sigma_high': np.nan}

            self._t0_mcmc = {'value': np.nan,
                    'sigma_low': np.nan, 
                    'sigma_high': np.nan}

            return np.nan, np.nan, np.nan, np.nan, np.nan, np.nan

        else:
            # process and log solution
            log_prob = sampler.get_log_prob(flat=True, discard=burnin)
            flat_samples = sampler.get_chain(flat=True, discard=burnin)
            flat_blobs = sampler.get_blobs(flat=True, discard=burnin)
            
            #figure out if there is branching in the solution and find the highest logp branch
            try:
                hist = np.histogram(log_prob, bins=50)
                arg_logp_max = np.argwhere(hist[0] != 0)[-1]
                logp_lim = hist[1][arg_logp_max-1]
                samples_top = flat_samples[log_prob >= logp_lim]
                blobs_top = flat_blobs[log_prob >= logp_lim]
                # log_prob_mean = np.mean(log_prob[log_prob >= logp_lim])
            except:
                samples_top = flat_samples
                blobs_top = flat_blobs
                # log_prob_mean = np.mean(log_prob)
            
            ndim = samples_top.shape[1]
            solution = []

            for j in range(ndim):
                mcmc = np.nanpercentile(samples_top[:, j], [16, 50, 84])
                q = np.diff(mcmc)
                solution.append([mcmc[1], q[0], q[1]])

            solution = np.array(solution)

            # compute the new period and errs
            means = solution[:,0]
            sigmas_low = solution[:,1]
            sigmas_high = solution[:,2]

            period = means[0]
            period_sigma_low = sigmas_low[0]
            period_sigma_high = sigmas_high[0]
            
            self._period_mcmc = {'value': period,
                                'sigma_low': period_sigma_low, 
                                'sigma_high': period_sigma_high}

            # compute the blob parameters
            # phasemin, residuals_mean, residuals_stdev, ecl1_area, ecl2_area
            ndim_blobs = blobs_top.shape[1]

            solution_blobs = []
            for j in range(ndim_blobs):
                mcmc_blob = np.nanpercentile(blobs_top[:, j], [16, 50, 84])
                q_blob = np.diff(mcmc_blob)
                solution_blobs.append([mcmc_blob[1], q_blob[0], q_blob[1]])

            solution_blobs = np.array(solution_blobs)
            means_blobs = solution_blobs[:,0]
            sigmas_low_blobs = solution_blobs[:,1]
            sigmas_high_blobs = solution_blobs[:,2]

            phasemin_mean = means_blobs[0]
            phasemin_sigma_low = sigmas_low_blobs[0]
            phasemin_sigma_high = sigmas_high_blobs[0]


            if np.isnan(self._t0_init):
                t0_new = 0 + period*phasemin_mean + int((self.__times.min()/period)+1)*(period)
            else:
                t0_new = self._t0_init + period*phasemin_mean

            t0_sigma_low = (period**2 * phasemin_sigma_low**2 + period_sigma_low**2 * phasemin_mean**2)**0.5
            t0_sigma_high = (period**2 * phasemin_sigma_high**2 + period_sigma_high**2 * phasemin_mean**2)**0.5

            self._t0_mcmc = {'value': t0_new,
                            'sigma_low': t0_sigma_low, 
                            'sigma_high': t0_sigma_high}

            return means, sigmas_low, sigmas_high, means_blobs, sigmas_low_blobs, sigmas_high_blobs


    def compute_model(self, means, sigmas_low, sigmas_high, save_lc = True, save_file='', show=False, failed=False):
        '''
        Computes the model parameter values from the sample.

        Parameters
        ----------
        means: array-like
            Mean values from the sample
        sigmas_low: array-like
            Standard deviation of samples < mean
        sigmas_high: array_like
            Standard deviation of samples > mean
        save_lc: bool
            If True, saves the model light curve to a file
        save_file: str
            File name to save light curve to, if save_lc=True.
        show: bool
            If True, will display a plot of the model light curve.
        failed: bool
            If True, all model parameters are np.nan
        '''
        # overriden by subclass
        return None


    def compute_eclipse_params(self, means, sigmas_low, sigmas_high, failed=False):
        '''
        Computes the model parameter values from the sample.

        Parameters
        ----------
        means: array-like
            Mean values from the sample
        sigmas_low: array-like
            Standard deviation of samples < mean
        sigmas_high: array_like
            Standard deviation of samples > mean
        failed: bool
            If true, all eclipse parameters are np.nan
        '''
        # pos1, width1, depth1, pos2, width2, depth2, ecl1_area, ecl2_area, residuals_mean, residuals_stdev
        eclipse_params = {'pos1': np.nan, 'width1': np.nan, 'depth1': np.nan, 
                          'pos2': np.nan, 'width2': np.nan, 'depth2': np.nan,
                          'ecl1_area': np.nan, 'ecl2_area': np.nan, 
                          'residuals_mean': np.nan, 'residuals_stdev': np.nan
                        }

        eclipse_params_err = {'pos1': np.nan, 'width1': np.nan, 'depth1': np.nan, 
                          'pos2': np.nan, 'width2': np.nan, 'depth2': np.nan,
                          'ecl1_area': np.nan, 'ecl2_area': np.nan, 
                          'residuals_mean': np.nan, 'residuals_stdev': np.nan
                        }

        if not failed:
            for ind, eclkey in enumerate(eclipse_params.keys()):
                # print(eclkey, means[ind], np.max((sigmas_low[ind],sigmas_high[ind])))
                eclipse_params[eclkey] = means[ind]
                eclipse_params_err[eclkey] = np.max((sigmas_low[ind],sigmas_high[ind]))

        
        self._eclipse_params = eclipse_params
        self._eclipse_params_errs = eclipse_params_err
        

    def save_results_to_file(self, results_file, type='ephemerides', ind=''):
        '''
        Save the resulting ephemerides, model or eclipse parameters to a file.

        Parameters
        ----------
        results_file: str
            Filename to save to
        type: str
            Which parameters to store. Available choices: ['ephemerides', 'model_values', 'eclipse_parameters']
        ind: str
            Index of the object (if looping through a list, otherwise optional)
        '''

        if type=='ephemerides':
            with open(results_file, 'a') as f:
                f.write('{},{},{},{},{},{},{},{}\n'.format(ind,
                                                            self._period_mcmc['value'],
                                                            self._period_mcmc['sigma_low'],
                                                            self._period_mcmc['sigma_high'],
                                                            self._t0_mcmc['value'],
                                                            self._t0_mcmc['sigma_low'],
                                                            self._t0_mcmc['sigma_high'],
                                                            self._chi2))

        elif type=='model_values':
            results_str = self._func+','

            for i,mkey in enumerate(self._model_values.keys()):
                results_str += '{},{}'.format(self._model_values[mkey],self._model_values_errs[mkey])
                if i < len(self._model_values.keys())-1:
                    results_str += ','
                else:
                    results_str += '\n'

            with open(results_file, 'a') as f:
                f.write('{},{}'.format(ind,results_str))


        elif type == 'eclipse_parameters':
            with open(results_file, 'a') as f:
                # pos1, pos1_err, width1, width1_err, depth1, depth1_err, 
                # pos2, pos2_err, width2, width2_err, depth2, depth2_err, 
                # res_mean, res_mean_err, res_stdev, res_stdev_err, 
                # ecl1_area, ecl1_area_err, ecl2_area, ecl2_area_err
                ecl = self._eclipse_params
                eclerr = self._eclipse_params_errs
                f.write('{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{}\n'.format(ind,
                    ecl['pos1'], eclerr['pos1'], ecl['width1'], eclerr['width1'], ecl['depth1'], eclerr['depth1'],
                    ecl['pos2'], eclerr['pos2'], ecl['width2'], eclerr['width2'], ecl['depth2'], eclerr['depth2'],
                    ecl['residuals_mean'], eclerr['residuals_mean'], ecl['residuals_stdev'], eclerr['residuals_stdev'],
                    ecl['ecl1_area'], eclerr['ecl1_area'], ecl['ecl2_area'], eclerr['ecl2_area']
                ))

        else:
            raise NotImplementedError


import numpy as np
from numpy import sin, cos, pi
from scipy.optimize import newton, minimize

class EbParams(object):

    def __init__(self, eclipse_params, fit_eclipses=True, **kwargs):

        '''
        Computes estimates of ecc, w, rsum and teffratio based on the eclipse parameters.

        Parameters
        ----------
        eclipse_params: dict
            Dictionary of the eclipse parameters determined from the two-Gaussian model or manually.
        refine_with_ellc: bool
            If true, an ellc.lc model will be fitted to the eclipses only to further refine 
            rsum, teffratio, as well as rratio and incl.

        '''

        self.pos1 = eclipse_params['primary_position']
        self.pos2 = eclipse_params['secondary_position']
        self.width1 = eclipse_params['primary_width']
        self.width2 = eclipse_params['secondary_width']
        self.depth1 = eclipse_params['primary_depth']
        self.depth2 = eclipse_params['secondary_depth']
        self.edges = eclipse_params['eclipse_edges']
         # computation fails if sep<0, so we need to adjust for it here.
        sep = self.pos2 - self.pos1
        if sep < 0:
            self.sep = 1+sep
        else:
            self.sep = sep

        self._ecc_w()
        self._teffratio()
        self._rsum()

        if fit_eclipses:
            phases = kwargs.get('phases', [])
            fluxes = kwargs.get('fluxes', [])
            sigmas = kwargs.get('sigmas', [])

            if len(phases) == 0 or len(fluxes) == 0 or len(sigmas) == 0:
                raise ValueError('Please provide values for the phases, fluxes and sigmas of the light curve!')

            self.refine_with_ellc(phases, fluxes, sigmas)
        else:
            self.rratio = 1.
            self.incl = 90.


    @staticmethod
    def _f (psi, sep): # used in pf_ecc_psi_w
        '''Returns the function to minimize for Psi'''
        return psi - sin(psi) - 2*pi*sep

    @staticmethod
    def _df (psi, sep): # used in pf_ecc_psi_w
        '''Returns the derivative of f for minimization'''
        return 1 - cos(psi) +1e-6

    def _ecc_w(self):
        '''Computes eccentricity and argument of periastron from the separation and widths.'''

        if np.isnan(self.sep) or np.isnan(self.width1) or np.isnan(self.width2):
            print('Cannot esimate eccentricty and argument of periastron: incomplete geometry information')
            return 0., np.pi/2
            
       
        psi = newton(func=self._f, x0=(12*np.pi*self.sep)**(1./3), fprime=self._df, args=(self.sep,), maxiter=5000)
        # ecc = sqrt( (0.25*(tan(psi-pi))**2+(swidth-pwidth)**2/(swidth+pwidth)**2)/(1+0.25*(tan(psi-pi))**2) )
        ecc = (np.sin(0.5*(psi-np.pi))**2+((self.width2-self.width1)/(self.width2+self.width1))**2*np.cos(0.5*(psi-np.pi))**2)**0.5
        try:
            w1 = np.arcsin((self.width1-self.width1)/(self.width2+self.width1)/ecc)
            w2 = np.arccos((1-ecc**2)**0.5/ecc * np.tan(0.5*(psi-np.pi)))

            w = w2 if w1 >= 0 else 2*pi-w2
        except:
            w = pi/2

        self.ecc = ecc
        self.per0 = w
        self.esinw = ecc*np.sin(w)
        self.ecosw = ecc*np.cos(w)


    def _t0_from_geometry(self, times, period=1, t0_supconj = 0, t0_near_times = True):
        '''
        Computes a new value for t0 from the position of the primary eclipse.

        Parameters
        ----------
        times: array-like
            Array of observed times
        period: float
            Orbital period of the object
        t0_supconj: float
            Initial t0 value (before fitting), if available. Default is 0.
        t0_near_times: bool
            If True, the computed t0 will be shifted to fall within the range of observed times.
        '''

        delta_t0 = self.pos1*period
        t0 = t0_supconj + delta_t0

        if t0_near_times:
            if t0 >= times.min() and t0 <= times.max():
                return t0
            else:
                return t0 + int((times.min()/period)+1)*(period)
        else:
            return t0
        

    def _teffratio(self):
        '''
        Computes the temprature ratio from eclipse depths.

        Holds only under the assumption of ecc=0, but it's the best first analytical guess we can get.
        '''
        self.teffratio = (self.depth2/self.depth1)**0.25

    
    def _rsum(self):
        '''
        Computes the sum of fractional radii from the eclipse widths, eccentricity and argument of periastron.

        The full equation for the eclipse widths contains a factor that depends on cos(incl) and requivratio.
        If we assume incl=90, that factor is 0 and the remaining is what we use to derive the equations for
        requivsum based on the widths of the primary and secondary eclipse below.
        '''
        rsum1 = np.pi*self.width1*(1-self.ecc**2)/(1+self.ecc*np.sin(self.per0))
        rsum2 = np.pi*self.width2*(1-self.ecc**2)/(1-self.ecc*np.sin(self.per0))
        self.rsum = np.average([rsum1, rsum2])


    def refine_with_ellc(self, phases, fluxes, sigmas):
        '''
        Refines the eclipse fits with an ellc light curve.

        Parameters
        ----------
        phases: array-like
            Orbital phases of the observed light curve
        fluxes: array-like
            Observed fluxes
        sigmas: array-like
            Flux uncertainities
        '''
        try:
            import ellc
        except:
            raise ImportError('ellc is required for parameter refinement, please install it before running this step.')
        
        def wrap_around_05(phases):
            phases[phases>0.5] = phases[phases>0.5] - 1
            phases[phases<-0.5] = phases[phases<-0.5] + 1
            return phases

        def mask_eclipses():

            edges = wrap_around_05(np.array(self.edges))
            ecl1 = edges[:2]
            ecl2 = edges[2:]

            if ecl1[1]>ecl1[0]:
                mask1 = (phases>=ecl1[0]) & (phases<=ecl1[1])
            else:
                mask1 = (phases>=ecl1[0]) | (phases<=ecl1[1])

            if ecl2[1]>ecl2[0]:
                mask2 = (phases>=ecl2[0]) & (phases<=ecl2[1])
            else:
                mask2 = (phases>=ecl2[0]) | (phases<=ecl2[1])

            phases_ecl, fluxes_ecl, sigmas_ecl = phases[mask1 | mask2], fluxes[mask1 | mask2], sigmas[mask1 | mask2]
            phases_outofecl, fluxes_outofecl, sigmas_outofecl = phases[~(mask1 | mask2)], fluxes[~(mask1 | mask2)], sigmas[~(mask1 | mask2)]

            meanf = np.mean(fluxes_outofecl) - 1
            return phases_ecl, fluxes_ecl, sigmas_ecl, meanf


        def lc_model(phases_mask, rsum, rratio, teffratio, incl, meanf):

            r1 = rsum/(1+rratio)
            r2 = rsum*rratio/(1+rratio)
            sbratio = np.sign(teffratio) * teffratio**4

            return ellc.lc(phases_mask, r1, r2, sbratio, incl, 
                light_3 = 0, 
                t_zero = self.pos1, period = 1,
                q = 1,
                f_c = self.ecc**0.5*np.cos(self.per0), f_s = self.ecc**0.5*np.sin(self.per0),
                shape_1='roche', shape_2='roche', 
                ld_1='lin', ld_2='lin', ldc_1=0.5, ldc_2=0.5,
                gdc_1=0., gdc_2=0., heat_1=0., heat_2=0.) + meanf

        def chi2(params, lc_data, meanf):
            rsum, rratio, teffratio, incl = params
            try:
                lcm = lc_model(lc_data[:,0], rsum, rratio, teffratio, incl, meanf)
                return 0.5 * np.sum((lc_data[:,1] - lcm) ** 2 / lc_data[:,2]**2)
            except Exception as e:
                return np.inf


        phases_mask, fluxes_mask, sigmas_mask, meanf = mask_eclipses()
        # phases_mask, fluxes_mask, sigmas_mask = phases, fluxes, sigmas
        lc_data = np.array([phases_mask, fluxes_mask, sigmas_mask]).T
        rsum_0 = self.rsum
        rratio_0 = 1.0
        teffratio_0 = self.teffratio 
        incl_0 = 90. 
        params_0 = [rsum_0, rratio_0, teffratio_0, incl_0]

        res = minimize(chi2, params_0, args=(lc_data, meanf), method='nelder-mead', options={'maxiter':10000})

        [self.rsum, self.rratio, self.teffratio, self.incl] = res.x


import numpy as np
from ligeor.fitting.sampler import EmceeSampler
from ligeor.models.polyfit import Polyfit
from ligeor.utils.lcutils import *


class EmceeSamplerPolyfit(EmceeSampler):
    
    def __init__(self, filename='', times = [], fluxes = [], sigmas = [],
                period_init=1, t0_init=0, n_downsample=0, **kwargs):
        '''
        Initializes a sampler for the light curve stored in 'filename'
        with determined initial values for the period and t0.

        Parameters
        ----------
        filename: str
            Filename to load the raw light curve from (expected format: times, fluxes, sigmas)
        times: array-like
            Time points of the observations
        fluxes: array-like
            Observed fluxes
        sigmas: array-like
            Uncertainities of the observed fluxes
        period_init: float
            Initial value of the period (from code/triage)
        t0_init: float
            Initial value of the time of superior conjunction (t0)
        n_downsample: int
            Number of data points to skip in raw light curve for downsampling
        nbins: int
            Number of bins (applies to the computed final model).
        
        Keyword arguments are passed on to load_lc.
            
        
        Attributes
        ----------
        _period_init: float
            Initial value of the orbital period
        _t0_init: float
            Initial value of the time of superior conjunction
        '''
        
        super(EmceeSamplerPolyfit,self).__init__(filename, times, fluxes, sigmas, period_init, t0_init, n_downsample, **kwargs)


    def initial_fit(self, knots = [], coeffs = []):
        '''
        Runs an initial fit to the data with the chosen model (two-Gaussian or polyfit).
        
        Attributes
        ----------
        _func: str
            Name of the best fit model function.
        _model_params: array-like
            Labels of all model parameters
        _initial_fit: array-like
            Initial values of the model parameters from optimization.
        '''
        
        if len(knots) == 0:
            phases, fluxes_ph, sigmas_ph = phase_fold(self._times, 
                                                            self._fluxes, 
                                                            self._sigmas, 
                                                            period=self._period_init, 
                                                            t0=self._t0_init, interval='01')

            polyfit = Polyfit(phases=phases, 
                                        fluxes=fluxes_ph, 
                                        sigmas=sigmas_ph, 
                                        )
            polyfit.fit()

            self._initial_fit = np.hstack((np.array(polyfit.knots), np.array(polyfit.coeffs).reshape(12)))

        else:
            if len(coeffs) == 0:
                phases, fluxes_ph, sigmas_ph = phase_fold(self._times, 
                                                            self._fluxes, 
                                                            self._sigmas, 
                                                            period=self._period_init, 
                                                            t0=self._t0_init, interval='01')

                polyfit = Polyfit(phases=phases, 
                                            fluxes=fluxes_ph, 
                                            sigmas=sigmas_ph, 
                                            )
                polyfit.fit(knots = knots)
                self._initial_fit = np.hstack((np.array(knots), np.array(polyfit.coeffs).reshape(12)))
            
            else:
                self._initial_fit = np.hstack((np.array(knots), np.array(coeffs).reshape(12)))

        self._func = 'PF'
        self._model_params = ['k1', 'k2', 'k3', 'k4', 'c11', 'c12', 'c13', 'c21', 'c22', 'c23', 'c31', 'c32', 'c33', 'c41', 'c42', 'c43']
        self._bounds = []
        for i,value in enumerate(self._initial_fit):
            self._bounds.append([value-0.1,value+0.1*value])


    def logprob(self, values):
        '''
        Computes the logprobability of the sample.

        Parameters
        ----------
        values: array-like
            period (for phase folding) + model values
        '''
        # bounds = [[-1e-5,-1e-5,-1e-5,-1e-5],[1+1e-5,1+1e-5,1+1e-5,1+1e-5]]
        # bounds = [[-1,-1,-1,-1],[2,2,2,2]]
        period, *model_vals = values

        for i,param_val in enumerate(model_vals[:4]):
            if param_val < self._bounds[i][0] or param_val > self._bounds[i][1]:
                # print('out of prior', self._bounds[i][0], self._bounds[i][1], param_val)
                return -np.inf, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan

        phases, fluxes_ph, sigmas_ph = phase_fold(self._times, 
                                                        self._fluxes, 
                                                        self._sigmas, 
                                                        period=period, 
                                                        t0=self._t0_init, interval='01')
        
        try:
            #TODO: figure out the cause for the "cannot unpack error"
            polyfit = Polyfit(phases=phases, 
                                        fluxes=fluxes_ph, 
                                        sigmas=sigmas_ph)
            polyfit.fit(knots = np.array(model_vals[:4]), coeffs = np.array(model_vals[4:]).reshape(4,3))
            
            eclipse_dir = polyfit.compute_eclipse_params()
            pos1, pos2 = eclipse_dir['primary_position'], eclipse_dir['secondary_position']
            width1, width2 = eclipse_dir['primary_width'], eclipse_dir['secondary_width']
            depth1, depth2 = eclipse_dir['primary_depth'], eclipse_dir['secondary_depth']
            ecl1_area, ecl2_area = polyfit.eclipse_area[1], polyfit.eclipse_area[2]
            residuals_mean, residuals_stdev = compute_residuals_stdev(fluxes_ph, polyfit.model)

            # print('residuals: ', residuals_mean, residuals_stdev)
            logprob = -0.5*(np.sum((fluxes_ph-polyfit.model)**2/sigmas_ph**2))
            # print(logprob, pos1, width1, depth1, pos2, width2, depth2)#, ecl1_area, ecl2_area, residuals_mean, residuals_stdev)
            return logprob, pos1, width1, depth1, pos2, width2, depth2, ecl1_area, ecl2_area, residuals_mean, residuals_stdev
        except:
            return -np.inf, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan


    def compute_model(self, means, sigmas_low, sigmas_high, save_lc = True, save_file='', show=False, failed=False):
        '''
        Computes the model parameter values from the sample.

        Parameters
        ----------
        means: array-like
            Mean values from the sample
        sigmas_low: array-like
            Standard deviation of samples < mean
        sigmas_high: array_like
            Standard deviation of samples > mean
        save_lc: bool
            If True, saves the model light curve to a file
        save_file: str
            File name to save light curve to, if save_lc=True.
        show: bool
            If True, will display a plot of the model light curve.
        failed: bool
            If True, all model parameters are np.nan
        '''
                
        model_results = {'k1': np.nan, 'k2': np.nan, 'k3': np.nan, 'k4': np.nan,
                        'c11': np.nan, 'c12': np.nan, 'c13': np.nan,
                        'c21': np.nan, 'c22': np.nan, 'c23': np.nan,
                        'c31': np.nan, 'c32': np.nan, 'c33': np.nan,
                        'c41': np.nan, 'c42': np.nan, 'c43': np.nan
                        }
        model_results_err = {'k1': np.nan, 'k2': np.nan, 'k3': np.nan, 'k4': np.nan,
                            'c11': np.nan, 'c12': np.nan, 'c13': np.nan,
                            'c21': np.nan, 'c22': np.nan, 'c23': np.nan,
                            'c31': np.nan, 'c32': np.nan, 'c33': np.nan,
                            'c41': np.nan, 'c42': np.nan, 'c43': np.nan
                        }
        if failed:
            self._model_values = model_results
            self._model_values_errs = model_results_err
            self._chi2 = np.nan

        else:
            # results_str = '{}'.format(func)
            for mkey in model_results.keys():
                if mkey in self._model_params:
                    pind = self._model_params.index(mkey)
                    model_results[mkey] = means[pind+1]
                    model_results_err[mkey] = np.max((sigmas_low[pind+1],sigmas_high[pind+1]))
                    # results_str += ',{},{}'.format(model_results[mkey],model_results_err[mkey])
            
            self._model_values = model_results
            self._model_values_errs = model_results_err
            chi2 = np.nan

            
            phases_obs, fluxes_ph_obs, sigmas_ph_obs = phase_fold(self._times, 
                                                    self._fluxes, 
                                                    self._sigmas, 
                                                    period=self._period_mcmc['value'], 
                                                    t0=self._t0_init,
                                                    interval='01')
        
            polyfit = Polyfit(phases=phases_obs, 
                                    fluxes=fluxes_ph_obs, 
                                    sigmas=sigmas_ph_obs)

            knots = np.array(list(self._model_values.values())[:4])
            coeffs = np.array(list(self._model_values.values())[4:]).reshape(4,3)
            polyfit.fit(knots = knots, coeffs = coeffs)
            fluxes_model = polyfit.fv(x = phases_obs)
            chi2 = -0.5*(np.sum((fluxes_ph_obs-fluxes_model)**2/sigmas_ph_obs**2))
            self._chi2 = chi2

            if show:
                import matplotlib.pyplot as plt
                plt.plot(phases_obs, fluxes_ph_obs, 'k.')
                plt.plot(phases_obs, fluxes_model, 'r-')
                plt.show()

            if save_lc:
                np.savetxt(save_file, np.array([phases_obs, fluxes_model]).T)


import numpy as np
import distl


def sample_skewed_gaussian(mean, sigma_low, sigma_high, size=1000):
    '''
    Samples from a skewed Gaussian distribution (hacky).
    
    Assumes different sigma left and right from the mean, samples them separately
    and combines into one sample, stitched together at the mean.
    
    Parameters
    ----------
    mean: float
        Mean of the Gaussian distribution
    sigma_low: float
        Standard deviation of the left Gaussian
    sigma_high: float
        Standard deviation of the right Gaussian
    size: int
        Number of samples to return
        
    Returns
    -------
    An array of the samples drawn from the distribution.
    '''
    
    samples_low = distl.gaussian(mean, sigma_low).sample(size)
    samples_high = distl.gaussian(mean, sigma_high).sample(size)
    
    samples_low_cut = samples_low[samples_low < mean]
    samples_high_cut = samples_high[samples_high >= mean]
    
    return np.hstack((samples_low_cut, samples_high_cut))

def combine_dists_from_hist(samples1, samples2, bins=1000, plot=False):
    '''
    Combine samples from two models and computes new mean and standard deviation.
    
    Parameters
    ----------
    samples1: array-like
        Samples drawn from first distribution.
    samples2: array-like
        Samples drawn from second distribution.
    bins: int
        Number of bins for the histogram
    plot: bool
        If True, will display the histogram.
        
    Returns
    -------
    A dictionary with the mean, sigma_low, sigma_high and confidence of the combined
    distribution.
    '''
    
    samples_full = np.hstack((samples1, samples2))
    hist_combined = distl.histogram_from_data(samples_full, bins=bins)
    
    if plot:
        hist_combined.plot()
        
    uncs = hist_combined.uncertainties()
    return {'mean': uncs[1], 
            'sigma_low': uncs[1]-uncs[0], 
            'sigma_high': uncs[2]-uncs[1],
            'confidence': 1}

def compute_combined_2g_pf_value(value_2g=1, value_2g_sigma_low=0, value_2g_sigma_high=0, weight_2g = 0.5,
                            value_pf=1, value_pf_sigma_low=0, value_pf_sigma_high=0, weight_pf = 0.5,
                            nsamples = 10000):
    '''
    Compute the combined value from two-Gaussian and polyfit MCMC results.
    
    Parameters
    ----------
    value_2g: float
        TwoGaussian value mean
    value_2g_sigma_low: float
        TwoGaussian value std for samples < value_2g
    value_2g_sigma_high: float
        TwoGaussian value std for samples > value_2g
    weight_2g: float
        The ratio of samples to be drawn from the TwoGaussian distribution, typically 1/chi2_2g
    value_pf: float
        Polyfit value mean
    value_pf_sigma_low: float
        Polyfit value std for samples < value_pf
    value_pf_sigma_high: float
        Polyfit value std for samples > value_pf
    weight_pf: float
        The ratio of samples to be drawn from the Polyfit distribution, typically 1/chi2_pf
    nsamples: int
        Number of samples for the combined distribution.   
        
    Returns
    -------
    value_mean: float
        The mean calculated from the combined distributions.
    value_sigma_low: float
        The standard deviation of the samples < mean
    value_sigma_high: float
        The standard deviation of the samples > mean 
    '''
    # from 2g and pf to combined values
        
    print('2g fit: P = %.5f + %.5f - %.5f' % (value_2g, value_2g_sigma_high, value_2g_sigma_low))
    print('pf fit: P = %.5f + %.5f - %.5f' % (value_pf, value_pf_sigma_high, value_pf_sigma_low))
    
    if ~np.isnan(weight_2g) and ~np.isnan(weight_pf):
        wratio = weight_2g/weight_pf

        nsamples_2g = int(wratio*nsamples/(1+wratio))
        nsamples_pf = int(nsamples/(1+wratio))

        print('values check: w_2g = %.6f, w_pf=%.6f, N_2g = %i, N_pf = %i'
         % (weight_2g, weight_pf, nsamples_2g, nsamples_pf))

        samples_2g = sample_skewed_gaussian(value_2g, value_2g_sigma_low, value_2g_sigma_high, size=nsamples_2g)
        samples_pf = sample_skewed_gaussian(value_pf, value_pf_sigma_low, value_pf_sigma_high, size=nsamples_pf)

        combined_result = combine_dists_from_hist(samples_2g, samples_pf, bins=1000, plot=False)
        return combined_result['mean'], combined_result['sigma_low'], combined_result['sigma_high']
        

    elif np.isnan(weight_2g) and ~np.isnan(weight_pf):
        return value_pf, value_pf_sigma_low, value_pf_sigma_high

        
    elif ~np.isnan(weight_2g) and np.isnan(weight_pf):
        return value_2g, value_2g_sigma_low, value_2g_sigma_high

    
    else:
        return np.nan, np.nan, np.nan


import numpy as np
from scipy.optimize import curve_fit
from ligeor.utils.lcutils import *
from ligeor.models import Model


class TwoGaussianModel(Model):

    def __init__(self, phases=[], fluxes=[], sigmas=[], filename='', 
                n_downsample=1, usecols=(0,1,2), delimiter=',', phase_folded=True, period=1, t0=0):
        '''
        Computes the two-Gaussian model light curves of the input data.

        Parameters
        ----------
        phases: array-like
            Input orbital phases, must be on the range [-0.5,0.5]
        fluxes: array-like
            Input fluxes
        sigmas: array-like
            Input sigmas (corresponding flux uncertainities)
        filename: str
            Filename from which to load a PHASE FOLDED light curve.
        n_downsample: int
            Number of data points to skip in loaded light curve (for downsampling)
        usecols: array-like, len 2 or 3
            Indices of the phases, fluxes and sigmas columns in file.
        '''

        super(TwoGaussianModel, self).__init__(phases, fluxes, sigmas, filename, n_downsample, usecols, delimiter, phase_folded, period, t0)

        # this is just all the parameter names for each model
        self.twogfuncs = {'C': TwoGaussianModel.const, 
                    'CE': TwoGaussianModel.ce, 
                    'CG': TwoGaussianModel.cg, 
                    'CGE': TwoGaussianModel.cge, 
                    'CG12': TwoGaussianModel.cg12, 
                    'CG12E1': TwoGaussianModel.cg12e1,
                    'CG12E2': TwoGaussianModel.cg12e2
                    }

        self.params = {'C': ['C'],
                'CE': ['C', 'Aell', 'phi0'],
                'CG': ['C', 'mu1', 'd1', 'sigma1'],
                'CGE': ['C', 'mu1', 'd1', 'sigma1', 'Aell', 'phi0'],
                'CG12': ['C', 'mu1', 'd1', 'sigma1', 'mu2', 'd2', 'sigma2'],
                'CG12E1': ['C', 'mu1', 'd1', 'sigma1', 'mu2', 'd2', 'sigma2', 'Aell'],
                'CG12E2': ['C', 'mu1', 'd1', 'sigma1', 'mu2', 'd2', 'sigma2', 'Aell']
                }


    def isolate_param_vals_for_model(self, model_params, func):
        '''
        Isolates only parameter values pertaining to a model "func",
        assuming model params contains the union of all possible parameters.
        
        Parameters
        ----------
        model_params: dict
            Dictionary containing all possible parameters and their values (some can be nan!)
            Keys: 'C', 'mu1', 'd1', 'sigma1', 'mu2', 'd2', 'sigma2', 'Aell', 'phi0'
            
        Returns
        -------
        param_vals: array-like
            A subset of the parameter values required to compute "func".
        '''
        param_vals = np.zeros(len(self.params[func]))
        for i,key in enumerate(self.params[func]):
            param_vals[i] = model_params[key]
            
        return param_vals


    def fit(self, fit_funcs=[], param_vals=[]):
        '''
        Computes all two-gaussian models and chooses the best fit.

        The fitting is carried out in the following steps: 
            1. fit each of the 7 models,
            2. compute their model light curves
            3. compute each model's BIC value
            4. assign the model with highest BIC as the 'best fit'
        '''
        self.fits = {}
        self.best_fit = {}
        
        if len(fit_funcs) != 0:
            # the best fit function has been provided by the user

            if len(param_vals) == 0:
                # we need to fit just that one
                self.fit_twoGaussian_models(funcs=fit_funcs)
            else:
                for i,key in enumerate(fit_funcs):
                    self.fits[key] = param_vals[i]
 
        else:
            self.fit_twoGaussian_models()
            # compute all model light curves

        self.compute_twoGaussian_models()
        # compute corresponding BIC values
        self.compute_twoGaussian_models_BIC()
        
        # best_fit_func = self.check_fit()
        # choose the best fit as the one with highest BIC
        best_fit_func = list(self.models.keys())[np.nanargmax(list(self.bics.values()))]
        self.best_fit['func'] = best_fit_func
        self.best_fit['param_vals'] = self.fits[best_fit_func]
        self.best_fit['param_names'] = self.params[best_fit_func]
            
        self.model = self.models[best_fit_func]


    def save_model(self, nbins=1000, func='', param_values = [], save_file=''):
        '''
        Save the best fit model to a file.

        Parameters
        ----------
        nbins: int
            The number of phase points between -0.5 and 0.5 to compute the model in
        save_file: str
            The filename to save to.
        '''
        
        if len(save_file) == 0:
            save_file = self.filename + '.2g'
        phases = np.linspace(-0.5,0.5,nbins)
        fluxes = self.twogfuncs[func](phases, *param_values)
        np.savetxt(save_file, np.array([phases, fluxes]).T)

    # HELPER FUNCTIONS

    @staticmethod
    def ellipsoidal(phi, Aell, phi0):
        '''
        Ellipsoidal model, defined as $y = (1/2) A_{ell} \cos (4 \pi (\phi - \phi_0))$
        
        Parameters
        ----------
        phi: float or array-like
            The input phase or phases array to compute the model in
        Aell: float
            Amplitude of the elliposoidal
        phi0: float
            Phase-point to center the elliposoidal on (position of primary or secondary eclipse)

        Returns
        -------
        y: array-like
            $y = (1/2) A_{ell} \cos (4 \pi (\phi - \phi_0))$
        '''
        # just the cosine component with the amplitude and phase offset as defined in Mowlavi (2017)
        return 0.5*Aell*np.cos(4*np.pi*(phi-phi0))

    @staticmethod
    def gaussian(phi, mu, d, sigma):
        '''
        Gaussian model, defined as $y = d \exp(-(\phi-\mu)^2/(2\sigma^2))$

        Parameters
        ----------
        phi: float or array-like
            The input phase or phases array to compute the model in
        mu: float
            Position of the Gaussian
        d: float
            Amplitude of the Gaussian
        sigma: float
            Scale (FWHM) of the Gaussian

        Returns
        -------
        y: array-like
            $y = d \exp(-(\phi-\mu)^2/(2\sigma^2))$
        '''

        # one Gaussian
        return d*np.exp(-(phi-mu)**2/(2*sigma**2))

    @staticmethod
    def gsum(phi, mu, d, sigma):
        '''
        Copies the Gaussian to mu +/- 1 and mu +/- 2 to account for 
        extended phase-folded light curves that cover multiple orbits.
        
        Parameters
        ----------
        phi: float or array-like
            The input phase or phases array to compute the model in
        mu: float
            Position of the Gaussian
        d: float
            Amplitude of the Gaussian
        sigma: float
            Scale (FWHM) of the Gaussian

        Returns
        -------
        y: array-like
            Array with gaussians at mu, mu +/-1 and mu +/- 2.
        '''

        gauss_sum = np.zeros(len(phi))
        for i in range(-2,3,1):
            gauss_sum += TwoGaussianModel.gaussian(phi,mu+i,d,sigma)
        return gauss_sum

    # MODELS as defined in Mowalvi (2017)

    @staticmethod
    def const(phi, C):
        '''The constant model y = C'
        
        Parameters
        ----------
        phi: float or array-like
            The input phase or phases array to compute the model in
        C: float
            value of the constant

        Returns
        -------
        y: array-like
            $y = C$
        '''

        # constant term
        return C*np.ones(len(phi))

    @staticmethod
    def ce(phi, C, Aell, phi0):
        '''Constant + ellipsoidal model
        
        Parameters
        ----------
        phi: float or array-like
            The input phase or phases array to compute the model in
        C: float
            value of the constant
        Aell: float
            Amplitude of the elliposoidal
        phi0: float
            Phase-point to center the elliposoidal on (position of primary or secondary eclipse)

        Returns
        -------
        y: array-like
            y = const(phi, C) - ellipsoidal(phi, Aell, phi0)
        '''

        return TwoGaussianModel.const(phi, C) - TwoGaussianModel.ellipsoidal(phi, Aell, phi0)

    @staticmethod
    def cg(phi, C, mu, d,  sigma):
        '''Constant + Gaussian model
        
        Parameters
        ----------
        phi: float or array-like
            The input phase or phases array to compute the model in
        C: float
            value of the constant
        mu: float
            Position of the Gaussian
        d: float
            Amplitude of the Gaussian
        sigma: float
            Scale (FWHM) of the Gaussian

        Returns
        -------
        y: array-like
            y = const(phi, C) - gsum(phi, mu, d, sigma)
        '''
        # constant + one gaussian (just primary eclipse)
        return TwoGaussianModel.const(phi, C) - TwoGaussianModel.gsum(phi, mu, d, sigma)

    @staticmethod
    def cge(phi, C, mu, d, sigma, Aell, phi0):
        '''
        Constant + Gaussian + ellipsoidal model

        Parameters
        ----------
        phi: float or array-like
            The input phase or phases array to compute the model in
        C: float
            value of the constant
        mu: float
            Position of the Gaussian
        d: float
            Amplitude of the Gaussian
        sigma: float
            Scale (FWHM) of the Gaussian
        Aell: float
            Amplitude of the elliposoidal
        phi0: float
            Phase-point to center the elliposoidal on (position of primary or secondary eclipse)

        Returns
        -------
        y: array-like
            y = const(phi, C) - gsum(phi, mu, d, sigma) - ellipsoidal(phi, Aell, mu)
        '''
        return TwoGaussianModel.const(phi, C) - TwoGaussianModel.ellipsoidal(phi, Aell, phi0) - TwoGaussianModel.gsum(phi, mu, d, sigma)

    @staticmethod
    def cg12(phi, C, mu1, d1, sigma1, mu2, d2, sigma2):
        '''
        Constant + two Gaussians model
        
        Parameters
        ----------
        phi: float or array-like
            The input phase or phases array to compute the model in
        C: float
            value of the constant
        mu1: float
            Position of the first Gaussian
        d1: float
            Amplitude of the first Gaussian
        sigma1: float
            Scale (FWHM) of the first Gaussian
        mu2: float
            Position of the second Gaussian
        d2: float
            Amplitude of the second Gaussian
        sigma2: float
            Scale (FWHM) of the second Gaussian

        Returns
        -------
        y: array-like
            y = const(phi, C) - gsum(phi, mu1, d1, sigma1) - gsum(phi, mu2, d2, sigma2) 
        '''
        return TwoGaussianModel.const(phi, C) - TwoGaussianModel.gsum(phi, mu1, d1, sigma1) - TwoGaussianModel.gsum(phi, mu2, d2, sigma2)

    @staticmethod
    def cg12e1(phi, C, mu1, d1, sigma1, mu2, d2, sigma2, Aell):
        '''
        Constant + two Gaussians + ellipsoidal centered on the primary eclipse
        
        Parameters
        ----------
        phi: float or array-like
            The input phase or phases array to compute the model in
        C: float
            value of the constant
        mu1: float
            Position of the first Gaussian
        d1: float
            Amplitude of the first Gaussian
        sigma1: float
            Scale (FWHM) of the first Gaussian
        mu2: float
            Position of the second Gaussian
        d2: float
            Amplitude of the second Gaussian
        sigma2: float
            Scale (FWHM) of the second Gaussian
        Aell: float
            Amplitude of the elliposoidal

        Returns
        -------
        y: array-like
            y = const(phi, C) - gsum(phi, mu1, d1, sigma1) - gsum(phi, mu2, d2, sigma2) - ellipsoidal(phi, Aell, mu1)
        '''
        return TwoGaussianModel.const(phi, C) - TwoGaussianModel.gsum(phi, mu1, d1, sigma1) - TwoGaussianModel.gsum(phi, mu2, d2, sigma2) - TwoGaussianModel.ellipsoidal(phi, Aell, mu1)

    @staticmethod
    def cg12e2(phi, C, mu1, d1, sigma1, mu2, d2, sigma2, Aell):
        '''
        Constant + two Gaussians + ellipsoidal centered on the secondary eclipse
        
        Parameters
        ----------
        phi: float or array-like
            The input phase or phases array to compute the model in
        C: float
            value of the constant
        mu1: float
            Position of the first Gaussian
        d1: float
            Amplitude of the first Gaussian
        sigma1: float
            Scale (FWHM) of the first Gaussian
        mu2: float
            Position of the second Gaussian
        d2: float
            Amplitude of the second Gaussian
        sigma2: float
            Scale (FWHM) of the second Gaussian
        Aell: float
            Amplitude of the elliposoidal
        phi0: float
            Phase-point to center the elliposoidal on (position of primary or secondary eclipse)


        Returns
        -------
        y: array-like
            y = const(phi, C) - gsum(phi, mu1, d1, sigma1) - gsum(phi, mu2, d2, sigma2) - ellipsoidal(phi, Aell, mu2)
        '''
        return TwoGaussianModel.const(phi, C) - TwoGaussianModel.gsum(phi, mu1, d1, sigma1) - TwoGaussianModel.gsum(phi, mu2, d2, sigma2) - TwoGaussianModel.ellipsoidal(phi, Aell, mu2)


    @staticmethod
    def lnlike(y, yerr, ymodel):
        '''
        Computes the loglikelihood of a model.

        $\log\mathrm{like} = \sum_i \log(\sqrt{2\pi} \sigma_i) + (y_i - model_i)^2/(2\sigma_i^2)
        '''
        if yerr is not None:
            return -np.sum(np.log((2*np.pi)**0.5*yerr)+(y-ymodel)**2/(2*yerr**2))
        else:
            return -np.sum((y-ymodel)**2)

    def bic(self, ymodel, nparams):
        '''
        Computes the Bayesian Information Criterion (BIC) value of a model.

        BIC = 2 lnlike - n_params \log(n_data)
        '''
        if self.sigmas is not None:
            return 2*self.lnlike(self.fluxes, self.sigmas, ymodel) - nparams*np.log(len(self.fluxes))
        else:
            return self.lnlike(self.fluxes, self.sigmas, ymodel)

    @staticmethod
    def find_eclipse(phases, fluxes):
        '''
        Determines the position and estimated width of the deepest eclipse.

        Parameters
        ----------
        phases: array-like
            Input phases of the light curve
        fluxes: array-like
            Input fluxes of the light curve

        Returns
        -------
        phase_min: float
            Position of the deepest eclipse in phase space
        edge_left: float
            Left edge of the eclipse in phase space
        edge_right: float
            Right edge of the eclipse in phase space
        '''

        phase_min = phases[np.nanargmin(fluxes)]
        ph_cross = phases[fluxes - np.nanmedian(fluxes) > 0]
        # this part looks really complicated but it really only accounts for eclipses split
        # between the edges of the phase range - if a left/right edge is not found, we look for
        # it in the phases on the other end of the range
        # we then mirror the value back on the side of the eclipse position for easier width computation
        try:
            arg_edge_left = np.argmin(np.abs(phase_min - ph_cross[ph_cross<phase_min]))
            edge_left = ph_cross[ph_cross<phase_min][arg_edge_left]
        except:
            arg_edge_left = np.argmin(np.abs((phase_min+1)-ph_cross[ph_cross<(phase_min+1)]))
            edge_left = ph_cross[ph_cross<(phase_min+1)][arg_edge_left]-1
        try:
            arg_edge_right = np.argmin(np.abs(phase_min-ph_cross[ph_cross>phase_min]))
            edge_right = ph_cross[ph_cross>phase_min][arg_edge_right]
        except:
            arg_edge_right = np.argmin(np.abs((phase_min-1)-ph_cross[ph_cross>(phase_min-1)]))
            edge_right = ph_cross[ph_cross>(phase_min-1)][arg_edge_right]+1

        return phase_min, edge_left, edge_right

    @staticmethod
    def estimate_eclipse_positions_widths(phases, fluxes, fix_primary=False, fix_secondary=False, diagnose_init=False):
        '''
        Determines the positions and widths of the eclipses in a light curve.

        Parameters
        ----------
        phases: array-like
            Input phases of the light curve
        fluxes: array-like
            Input fluxes of the light curve
        diagnose_init: bool
            If True, will plot the light curve with the position and width estimates.

        Returns
        -------
        estimates: dict
            A dictionary with keys 'ecl_positions' and 'ecl_widths' corresponding
            to lists of len 2 of the initial estimated values for each eclipse.
        '''
        
        if fix_primary:
            pos1 = phases[np.argmin(fluxes)]
            pos2 = pos1+0.5
            pos2 = pos2-1 if pos2 > 0.5 else pos2 
            return {'ecl_positions': [pos1, pos2], 'ecl_widths': [0.001, 0.001]}


        elif fix_secondary:
            pos1, edge1l, edge1r = TwoGaussianModel.find_eclipse(phases, fluxes)
            pos2 = pos1+0.5
            pos2 = pos2-1 if pos2 > 0.5 else pos2 
            return {'ecl_positions': [pos1, pos2], 'ecl_widths': [edge1r-edge1l, 0.005]}

        else:
            pos1, edge1l, edge1r = TwoGaussianModel.find_eclipse(phases, fluxes)
            fluxes_sec = fluxes.copy()
            fluxes_sec[((phases > edge1l) & (phases < edge1r)) | ((phases > edge1l+1) | (phases < edge1r-1))] = np.nan
            pos2, edge2l, edge2r = TwoGaussianModel.find_eclipse(phases, fluxes_sec)

            if diagnose_init:
                import matplotlib.pyplot as plt
                plt.figure(figsize=(10,8))
                plt.plot(phases, fluxes, '.')
                plt.axhline(y=np.median(fluxes), c='orange')
                for i,x in enumerate([pos1, edge1l, edge1r]):
                    ls = '-' if i==0 else '--'
                    plt.axvline(x=x, c='r', ls=ls)
                for i,x in enumerate([pos2, edge2l, edge2r]):
                    ls = '-' if i==0 else '--'
                    plt.axvline(x=x, c='g', ls=ls)
                plt.show()

            # return {'ecl_positions': [pos1, pos2], 'ecl_widths': [edge1r-edge1l, edge2r-edge2l]}
            return {'ecl_positions': [pos1, pos2], 'ecl_widths': [0.005, 0.005]}
        
    def fit_twoGaussian_models(self, init_pos=[], init_widths=[], 
                               funcs=['C', 'CE', 'CG', 'CGE', 'CG12', 'CG12E1', 'CG12E2']):
        '''
        Fits all seven models to the input light curve.
        '''
        # setup the initial parameters

        C0 = self.fluxes.max()
        ecl_dict = self.estimate_eclipse_positions_widths(self.phases, self.fluxes, diagnose_init=False)
        if len(init_pos) == 0 or len(init_widths) == 0:
            self.init_positions, self.init_widths = ecl_dict['ecl_positions'], ecl_dict['ecl_widths']
            mu10, mu20 = self.init_positions
            sigma10, sigma20 = self.init_widths
        else:
            mu10, mu20 = init_pos
            sigma10, sigma20 = init_widths

        # check if the initial estimates for the eclipses are overlapping
        if np.abs(mu20-mu10) < 5.6*sigma10 or np.abs(mu20-mu10) < 5.6*sigma20:
            mu20 = mu10 + 0.5
            mu20 = mu20 - 1 if mu20 > 0.5 else mu20

        d10 = self.fluxes.max()-self.fluxes[np.argmin(np.abs(self.phases-mu10))]
        d20 = self.fluxes.max()-self.fluxes[np.argmin(np.abs(self.phases-mu20))]
        Aell0 = 0.001
        phi0 = 0.25

        init_params = {'C': [C0,],
            'CE': [C0, Aell0, phi0],
            'CG': [C0, mu10, d10, sigma10],
            'CGE': [C0, mu10, d10, sigma10, Aell0, phi0],
            'CG12': [C0, mu10, d10, sigma10, mu20, d20, sigma20],
            'CG12E1': [C0, mu10, d10, sigma10, mu20, d20, sigma20, Aell0],
            'CG12E2': [C0, mu10, d10, sigma10, mu20, d20, sigma20, Aell0]}

        # parameters used frequently for bounds
        fmax = self.fluxes.max()
        fmin = self.fluxes.min()
        fdiff = fmax - fmin

        bounds = {'C': ((0),(fmax)),
            'CE': ((0, 1e-6, -0.5),(fmax, fdiff, 0.5)),
            'CG': ((0., -0.5, 0., 0.), (fmax, 0.5, fdiff, 0.5)),
            'CGE': ((0., -0.5, 0., 0., 1e-6, -0.5),(fmax, 0.5, fdiff, 0.5, fdiff, 0.5)),
            'CG12': ((0.,-0.5, 0., 0., -0.5, 0., 0.),(fmax, 0.5, fdiff, 0.5, 0.5, fdiff, 0.5)),
            'CG12E1': ((0.,-0.5, 0., 0., -0.5, 0., 0., 1e-6),(fmax, 0.5, fdiff, 0.5, 0.5, fdiff, 0.5, fdiff)),
            'CG12E2': ((0.,-0.5, 0., 0., -0.5, 0., 0., 1e-6),(fmax, 0.5, fdiff, 0.5, 0.5, fdiff, 0.5, fdiff))}


        # extend light curve on phase range [-1,1]
        phases_ext, fluxes_ext, sigmas_ext = extend_phasefolded_lc(self.phases, self.fluxes, self.sigmas)

        for key in funcs:
            try:
                self.fits[key] = curve_fit(self.twogfuncs[key], phases_ext, fluxes_ext, p0=init_params[key], sigma=sigmas_ext, bounds=bounds[key])[0]
            except Exception as err:
                print("2G model {} failed with error: {}".format(key, err))
                self.fits[key] = np.array([np.nan*np.ones(len(init_params[key]))])


    def compute_twoGaussian_models(self):
        '''
        Computes the model light curves given the fit solutions.
        '''
        models = {}

        for fkey in self.fits.keys():
            models[fkey] = self.twogfuncs[fkey](self.phases, *self.fits[fkey])

        self.models = models


    def compute_twoGaussian_models_BIC(self):
        '''
        Computes the BIC value of each model light curve.
        '''
        bics = {}
        nparams = {'C':1, 'CE':3, 'CG':4, 'CGE':6, 'CG12':7, 'CG12E1':8, 'CG12E2':8}

        for mkey in self.models.keys():
            bics[mkey] = self.bic(self.models[mkey], nparams[mkey])

        self.bics = bics


    def compute_eclipse_params(self, interactive=False):
        '''
        Compute the positions, widths and depths of the eclipses 
        based on the two-Gaussian model solution.

        The eclipse parameters are computed following the prescription
        in Mowlavi et al. (2017):
        - eclipse positions are set to the Gaussian positions
        - eclipse widths are 5.6 times the Gaussian FWHMs
        - eclipse depths are the constant factor minus the value of 
        the model at the eclipse positions

        Parameters
        ----------
        interactive: boolean
            If True, allows the user to manually adjust the positions of 
            eclipse minimum and edges. Default: False.

        Returns
        -------
        results: dict
            A dictionary of the eclipse paramter values.
        '''

        param_vals = self.best_fit['param_vals']
        param_names = self.best_fit['param_names']

        # gather values from the best fit solution
        sigma1 = param_vals[param_names.index('sigma1')] if 'sigma1' in param_names else np.nan
        sigma2 = param_vals[param_names.index('sigma2')] if 'sigma2' in param_names else np.nan
        mu1 = param_vals[param_names.index('mu1')] if 'mu1' in param_names else np.nan
        mu2 = param_vals[param_names.index('mu2')] if 'mu2' in param_names else np.nan
        C = param_vals[param_names.index('C')]

        # compute and adjust all available parameters, otherwise the entire eclipse is nan
        if not np.isnan(mu1) and not np.isnan(sigma1) and np.abs(sigma1) < 0.5:
            pos1 = mu1
            width1 = 5.6*np.abs(sigma1)
            depth1 = C - self.twogfuncs[self.best_fit['func']]([pos1], *self.best_fit['param_vals'])[0]
        else:
            pos1 = np.nan
            width1 = np.nan
            depth1 = np.nan
        if not np.isnan(mu2) and not np.isnan(sigma2) and np.abs(sigma2) < 0.5:
            pos2 = mu2
            width2 = 5.6*np.abs(sigma2)
            depth2 = C - self.twogfuncs[self.best_fit['func']]([pos2], *self.best_fit['param_vals'])[0]
        else:
            pos2 = np.nan
            width2 = np.nan
            depth2 = np.nan


        # compute the eclipse edges using the positons and widths
        eclipse_edges = [pos1 - 0.5*width1, pos1+0.5*width1, pos2-0.5*width2, pos2+0.5*width2]

        # # check if the resulting eclipses are overlapping:
        # if np.abs(pos2-pos1) < width1 or np.abs(pos2-pos1) < width2:
        #     # keep only the deeper eclipse
        #     if depth1 > depth2:
        #         pos1, pos2 = pos1, np.nan 
        #         width1, width2 = width1, np.nan 
        #         depth1, depth2 = depth1, np.nan
        #         eclipse_edges = [eclipse_edges[0], eclipse_edges[1], np.nan, np.nan]
        #     else:
        #         pos1, pos2 = pos2, np.nan 
        #         width1, width2 = width2, np.nan 
        #         depth1, depth2 = depth2, np.nan
        #         eclipse_edges = [eclipse_edges[2], eclipse_edges[3], np.nan, np.nan]

        self.eclipse_params = {
            'primary_width': width1,
            'secondary_width': width2,
            'primary_position': pos1,
            'secondary_position': pos2,
            'primary_depth': depth1,
            'secondary_depth': depth2,
            'eclipse_edges': eclipse_edges
        }

        if interactive:
           self.interactive_eclipse()
        self.compute_eclipse_area(ecl=1)
        self.compute_eclipse_area(ecl=2)
        # self.eclipse_params = self.check_eclipses_credibility()

        # check if eclipses need to be swapped:
        if ~np.isnan(self.eclipse_params['primary_depth']) and ~np.isnan(self.eclipse_params['secondary_depth']):
            if self.eclipse_params['secondary_depth'] > self.eclipse_params['primary_depth']:
                pos1, d1, w1, edge1 = self.eclipse_params['secondary_position'], self.eclipse_params['secondary_depth'], self.eclipse_params['secondary_width'], self.eclipse_params['eclipse_edges'][2:]
                pos2, d2, w2, edge2 = self.eclipse_params['primary_position'], self.eclipse_params['primary_depth'], self.eclipse_params['primary_width'], self.eclipse_params['eclipse_edges'][:2]

                self.eclipse_params['primary_position'] = pos1 
                self.eclipse_params['primary_width'] = w1
                self.eclipse_params['primary_depth'] = d1
                self.eclipse_params['secondary_position'] = pos2 
                self.eclipse_params['secondary_width'] = w2
                self.eclipse_params['secondary_depth'] = d2
                self.eclipse_params['eclipse_edges'] = [edge1[0],edge1[1], edge2[0], edge2[1]]
        return self.eclipse_params

    @staticmethod
    def compute_gaussian_area(mu, sigma, d, xmin, xmax):
        '''
        Computes the area under an eclipse.

        The area is limited to an x-range of of [-2.8*sigma, 2.8*sigma] to correspond to
        the definition of an eclipse width of 5.6*.

        Parameters
        ----------
        mu: float
            Position (mean) of the Gaussian
        sigma: float
            Standard deviation of the Gaussian
        d: float
            Depth (amplitude) of the Gaussian
        xmin: float
            Bottom integral boundary
        xmax: float
            Top integral boundary
        
        Returns
        -------
        area: float
            The computed area under the Gaussian.

        '''
        if np.isnan(mu) or np.isnan(sigma) or np.isnan(d):
            return np.nan
        else:
            from scipy import special

            integral_top = special.erf((xmax - mu)/(np.sqrt(2)*sigma))
            integral_bottom = special.erf((xmin - mu)/(np.sqrt(2)*sigma))
            return np.sqrt(np.pi/2)*d*sigma*(integral_top-integral_bottom)


    def compute_eclipse_area(self, ecl=1):
        '''
        Computes the area under an eclipse.

        An eclipse is defined as being positioned at the Gaussian postion (mu),
        with a depth corresponding to the Gaussian amplitude (d) and width of 5.6*sigma.

        Parameters
        ----------
        ecl: int
            The eclipse whose area is to be computed (1 or 2)

        '''

        if hasattr(self, 'eclipse_area'):
            pass
        else:
            self.eclipse_area = {}

        if ecl==1:
            if self.best_fit['func'] in ['C', 'CE']:
                self.eclipse_area[ecl] = np.nan
            else:
                mu_ind = self.best_fit['param_names'].index('mu%i' % ecl)
                sigma_ind = self.best_fit['param_names'].index('sigma%i' % ecl)
                d_ind = self.best_fit['param_names'].index('d%i' % ecl)
                
                mu = self.best_fit['param_vals'][mu_ind]
                sigma = self.best_fit['param_vals'][sigma_ind]
                d = self.best_fit['param_vals'][d_ind]
                phi_top = mu + 2.8*sigma
                phi_bottom = mu - 2.8*sigma

                self.eclipse_area[ecl] = TwoGaussianModel.compute_gaussian_area(mu, sigma, d, phi_top, phi_bottom)
            
        else:
            if self.best_fit['func'] in ['C', 'CE', 'CG', 'CGE']:
                self.eclipse_area[ecl] = np.nan
            else:
                mu_ind = self.best_fit['param_names'].index('mu%i' % ecl)
                sigma_ind = self.best_fit['param_names'].index('sigma%i' % ecl)
                d_ind = self.best_fit['param_names'].index('d%i' % ecl)
                
                mu = self.best_fit['param_vals'][mu_ind]
                sigma = self.best_fit['param_vals'][sigma_ind]
                d = self.best_fit['param_vals'][d_ind]
                phi_top = mu + 2.8*sigma
                phi_bottom = mu - 2.8*sigma

                self.eclipse_area[ecl] = TwoGaussianModel.compute_gaussian_area(mu, sigma, d, phi_top, phi_bottom)


    def plot(self):
        import matplotlib.pyplot as plt

        plt.plot(self.phases, self.fluxes, 'b.')
        plt.plot(self.phases, self.model, 'r-')

        for i in range(4):
            plt.axvline(self.eclipse_params['eclipse_edges'][i])

        plt.show()





class DraggableLine:

    def __init__(self, p):
        '''
        Enables manual adjustment of line positions corresponding to
        the eclipse positions and edges.

        Parameters
        ----------
        p: matplotlib ax.axvline component to be adjusted
        '''
        self.point = p
        self.press = None

    def connect(self):
        self.cidpress = self.point.figure.canvas.mpl_connect('button_press_event', self.button_press_event)
        self.cidrelease = self.point.figure.canvas.mpl_connect('button_release_event', self.button_release_event)
        self.cidmotion = self.point.figure.canvas.mpl_connect('motion_notify_event', self.motion_notify_event)

    def disconnect(self):
        #disconnect all the stored connection ids
        self.point.figure.canvas.mpl_disconnect(self.cidpress)
        self.point.figure.canvas.mpl_disconnect(self.cidrelease)
        self.point.figure.canvas.mpl_disconnect(self.cidmotion)


    def button_press_event(self,event):
        if event.inaxes != self.point.axes:
            return
        contains = self.point.contains(event)[0]
        if not contains: return
        self.press = self.point.get_xdata(), event.xdata

    def button_release_event(self,event):
        self.press = None
        self.point.figure.canvas.draw()

    def motion_notify_event(self, event):
        if self.press is None: return
        if event.inaxes != self.point.axes: return
        xdata, xpress = self.press
        dx = event.xdata-xpress
        self.point.set_xdata(xdata+dx)
        self.point.figure.canvas.draw()



import numpy as np

def load_lc(lc_file, n_downsample=0, phase_folded=False, usecols=(0,1,2), delimiter=','):
    '''
    Loads the light curve from lc_file and returns 
    separate arrays for times, fluxes and sigmas.

    Parameters
    ----------
    lc_file: str
        Filename to load light curve from.
    n_downsample: int
        Number of data points to skip in downsampling the lc.

    Returns
    -------
    A dictionary of the times, fluxes and sigmas retrieved from the lc file.
    '''
    
    lc = np.loadtxt(lc_file, usecols=usecols, delimiter=delimiter)

    if phase_folded:
        return {'phases': lc[:,0][::n_downsample], 
        'fluxes': lc[:,1][::n_downsample], 
        'sigmas': lc[:,2][::n_downsample]}

    else:
        return {'times': lc[:,0][::n_downsample], 
                'fluxes': lc[:,1][::n_downsample], 
                'sigmas': lc[:,2][::n_downsample]}


def phase_fold(times, fluxes, sigmas, period=1, t0=0, interval='05'):
    '''
    Phase-folds the light curve with a given period and t0.

    Parameters
    ----------
    times: array-like
        The observation times
    fluxes: array-like
        Observed fluxes corresponding to each time in times
    sigmas: array-like
        Uncertainties corresponding to each flux in fluxes
    interval: str
        Phase range to be returned, [-0.5,0.5] if interval='05' or [0.,1.] if interval='01'.


    Returns
    -------
    phases: array-like
        The computed orbital phases on a range [-0.5,0.5], sorted in ascending order
    fluxes_ph: array_like
        The fluxes resorted to match the phases array order
    sigmas_ph: array_like
        The sigmas resorted to match the phases array order
    '''
    
    t0 = 0 if np.isnan(t0) else t0
    phases = np.mod((times-t0)/period, 1.0)

    if interval == '05':
        if isinstance(phases, float):
            if phases > 0.5:
                phases -= 1
        else:
            # then should be an array
            phases[phases > 0.5] -= 1
        
    s = phases.argsort()
    phases = phases[s]
    fluxes_ph = fluxes[s]
    sigmas_ph = sigmas[s]
    
    return phases, fluxes_ph, sigmas_ph


def extend_phasefolded_lc(phases, fluxes, sigmas):
    '''
    Takes a phase-folded light curve on the range [-0.5,0.5] and extends it on range [-1,1]
    
    Parameters
    ----------
    phases: array-like
        Array of input phases spanning the range [-0.5,0.5]
    fluxes: array-like
        Corresponding fluxes, length must be equal to that of phases
    sigmas: array-like
        Corresponsing measurement uncertainties, length must be equal to that og phases
        
    Returns
    -------
    phases_extend, fluxes_extend, sigmas_extend: array-like
        Extended arrays on phase-range [-1,1]
    
    '''
    #make new arrays that would span phase range -1 to 1:
    fluxes_extend = np.hstack((fluxes[(phases > 0)], fluxes, fluxes[phases < 0.]))
    phases_extend = np.hstack((phases[phases>0]-1, phases, phases[phases<0]+1))

    if sigmas is not None:
        sigmas_extend = np.hstack((sigmas[phases > 0], sigmas, sigmas[phases < 0.]))
    else:
        sigmas_extend = None

    return phases_extend, fluxes_extend, sigmas_extend


def check_eclipse_fitting_noise(fluxes_model, fluxes_obs, depth):
    '''
    Checks if the model eclipses are true or fit data noise features.

    Parameters
    ----------
    fluxes_model: array-like
        Fluxes computed from the model
    fluxes_obs: array-like
        Observed fluxes
    depth: float
        Depth of the eclipse

    Returns
    -------
    bool
    '''

    sigma_res = np.std(fluxes_obs - fluxes_model)
    if depth <= 0.5*sigma_res:
        return True
    else:
        return False


def check_eclipse_fitting_cosine(width):
    '''
    Checks if an eclipse is fitted to the out-of-eclipse variability.

    Parameters
    ----------
    width: float
        Measured width of the eclipse from the model.

    Returns
    -------
    bool
    '''
    if width >= 0.41:
        return True
    else:
        return False

def check_eclipse_depth_negative(depth):
    '''
    Checks if an eclipse is fitting another feature that does not show a decrease in flux.
    Parameters
    ----------
    depth: float
        Measured depth of the eclipse from the model.

    Returns
    -------
    bool
    '''
    if depth <= 0.0:
        return True
    else:
        return False


def compute_residuals_stdev(fluxes_obs, fluxes_model):
    '''
    Computes the residuals of the input fluxes and best fit model

    Parameters
    ----------
    fluxes_obs: array-like
        Observed fluxes
    fluxes_model: array-like
        Model fluxes

    Returns
    -------
    residuals_mean: float
        The mean of the residuals
    residuals_stdev: float
        The standard deviation of the residuals
    '''
    residuals = fluxes_obs - fluxes_model
    return np.mean(residuals), np.std(residuals)


from os import PRIO_PGRP
import numpy as np
from ligeor.utils.lcutils import *

class Model(object):

    def __init__(self, phases, fluxes, sigmas, filename, n_downsample, usecols, delimiter, phase_folded, period, t0):
        '''
        Computes the two-Gaussian model light curves of the input data.

        Parameters
        ----------
        phases: array-like
            Input orbital phases, must be on the range [-0.5,0.5]
        fluxes: array-like
            Input fluxes
        sigmas: array-like
            Input sigmas (corresponding flux uncertainities)
        filename: str
            Filename from which to load a PHASE FOLDED light curve.
        n_downsample: int
            Number of data points to skip in loaded light curve (for downsampling)
        usecols: array-like, len 2 or 3
            Indices of the phases, fluxes and sigmas columns in file.
        '''

        if len(phases) == 0 or len(fluxes) == []:
            try:
                lc = load_lc(filename, n_downsample=n_downsample, phase_folded=phase_folded, usecols=usecols, delimiter=delimiter)
                if phase_folded:
                    self.phases = lc['phases']
                    self.fluxes = lc['fluxes']
                    self.sigmas = lc['sigmas']
                else:
                    self.phases, self.fluxes, self.sigmas = phase_fold(lc['times'], lc['fluxes'], lc['sigmas'], period=period, t0=t0)

            except Exception as e:
                raise ValueError(f'Loading light curve failed with exception {e}')
        else:
            self.filename = filename
            self.phases = phases 
            self.fluxes = fluxes 
            self.sigmas = sigmas


    def interactive_eclipse(self):
        '''
        Displays an interactive plot with draggable lines for the eclipse edges and positions.
        '''

        from ligeor.utils.interactive import DraggableLine
        phases_w, fluxes_w, sigmas_w = extend_phasefolded_lc(self.phases, self.fluxes, self.sigmas)
        [ecl1_l, ecl1_r, ecl2_l, ecl2_r] = self.eclipse_params['eclipse_edges']
        pos1, pos2 = self.eclipse_params['primary_position'], self.eclipse_params['secondary_position']
        
        import matplotlib.pyplot as plt
        fig = plt.figure(figsize=(10,8))
        ax = fig.add_subplot(111)
        ax.plot(phases_w, fluxes_w, 'k.')
        plt.plot(phases_w, self.twogfuncs[self.best_fit](phases_w, *self.fits[self.best_fit][0]), '-', label=self.best_fit)
        lines = []
        lines.append(ax.axvline(x=pos1, c='#2B71B1', lw=2, label='primary'))
        lines.append(ax.axvline(x=pos2, c='#FF702F', lw=2, label='secondary'))
        lines.append(ax.axvline(x=ecl1_l, c='#2B71B1', lw=2, ls='--'))
        lines.append(ax.axvline(x=ecl1_r, c='#2B71B1', lw=2, ls='--'))
        lines.append(ax.axvline(x=ecl2_l, c='#FF702F', lw=2, ls='--'))
        lines.append(ax.axvline(x=ecl2_r, c='#FF702F', lw=2, ls='--'))
        drs = []
        for l,label in zip(lines,['pos1', 'pos2', 'ecl1_l', 'ecl1_r', 'ecl2_l', 'ecl2_r']):   
            dr = DraggableLine(l)
            dr.label = label
            dr.connect()   
            drs.append(dr) 
        ax.legend()
        plt.show(block=True)

        print('adjusting values')

        pos1 = drs[0].point.get_xdata()[0]
        pos2 = drs[1].point.get_xdata()[0]
        ecl1_l = drs[2].point.get_xdata()[0]
        ecl1_r = drs[3].point.get_xdata()[0]
        ecl2_l = drs[4].point.get_xdata()[0]
        ecl2_r = drs[5].point.get_xdata()[0]

        width1 = ecl1_r - ecl1_l
        width2 = ecl2_r - ecl2_l

        #TODO: replace max with mean out of eclipses (or at edges)
        depth1 = self.fluxes.max() - self.fluxes[np.argmin(np.abs(self.phases-pos1))]
        depth2 = self.fluxes.max() - self.fluxes[np.argmin(np.abs(self.phases-pos2))]
        
        eclipse_edges = [ecl1_l, ecl1_r, ecl2_l, ecl2_r]

        self.eclipse_params = {
            'primary_width': width1,
            'secondary_width': width2,
            'primary_position': pos1,
            'secondary_position': pos2,
            'primary_depth': depth1,
            'secondary_depth': depth2,
            'eclipse_edges': eclipse_edges
    }



    def check_eclipses_credibility(self):
        '''
        Checks if the detected eclipses are statistically significant.

        The checks performed are:
        - whether the eclipse is fitted to a noise feature. If True, the eclipse parameters are discarded.
        - whether the two detected eclipses overlap. If True, only the shallower eclipse is discarded.
        '''
        
        eclipse_params = self.eclipse_params.copy()
        if hasattr(self, 'eclipse_params'):

            # check secondary eclipse
            if ~np.isnan(eclipse_params['secondary_position']):
                if (
                check_eclipse_fitting_noise(self.model, self.fluxes, eclipse_params['secondary_depth']) or 
                check_eclipse_fitting_cosine(eclipse_params['secondary_width'])):
                    print('SECONDARY ECLIPSE FITTING NOISE OR COSINE')
                    eclipse_params['secondary_position'] = np.nan 
                    eclipse_params['secondary_width'] = np.nan 
                    eclipse_params['secondary_depth'] = np.nan
                    eclipse_params['eclipse_edges'][2] = np.nan
                    eclipse_params['eclipse_edges'][3] = np.nan

            # check primary eclipse
            if ~np.isnan(eclipse_params['primary_position']):
                if (
                check_eclipse_fitting_noise(self.model, self.fluxes, eclipse_params['primary_depth']) or 
                check_eclipse_fitting_cosine(eclipse_params['primary_width'])):
                    print('PRIMARY ECLIPSE FITTING NOISE OR COSINE')
                
                    if ~np.isnan(eclipse_params['secondary_position']):
                        print('replacing with secondary')
                        eclipse_params['primary_position'] = float(eclipse_params['secondary_position'])
                        eclipse_params['primary_width'] = float(eclipse_params['secondary_width'])
                        eclipse_params['primary_depth'] = float(eclipse_params['secondary_depth'])
                        eclipse_params['eclipse_edges'][0] = float(eclipse_params['eclipse_edges'][2])
                        eclipse_params['eclipse_edges'][1] = float(eclipse_params['eclipse_edges'][3])
                        eclipse_params['secondary_position'] = np.nan 
                        eclipse_params['secondary_width'] = np.nan 
                        eclipse_params['secondary_depth'] = np.nan
                        eclipse_params['eclipse_edges'][2] = np.nan
                        eclipse_params['eclipse_edges'][3] = np.nan
                    else:
                        print('replacing with nans!')
                        eclipse_params['primary_position'] = np.nan 
                        eclipse_params['primary_width'] = np.nan 
                        eclipse_params['primary_depth'] = np.nan
                        eclipse_params['eclipse_edges'][0] = np.nan
                        eclipse_params['eclipse_edges'][1] = np.nan
                    
            # check overlapping eclipses
            if ~np.isnan(eclipse_params['primary_position']) and ~np.isnan(eclipse_params['secondary_position']):
                pos_dist = np.abs(eclipse_params['primary_position'] - eclipse_params['secondary_position'])
                if  pos_dist < eclipse_params['primary_width'] or pos_dist < eclipse_params['secondary_width']:
                    # keep only the deeper eclipse
                    if eclipse_params['primary_depth'] > eclipse_params['secondary_depth']:
                        # keep only the primary eclipse
                        eclipse_params['secondary_position'] = np.nan 
                        eclipse_params['secondary_width'] = np.nan 
                        eclipse_params['secondary_depth'] = np.nan
                        eclipse_params['eclipse_edges'][2] = np.nan
                        eclipse_params['eclipse_edges'][3] = np.nan

                    else:
                        # keep the secondary as primary and discard secondary
                        eclipse_params['primary_position'] = float(eclipse_params['secondary_position'])
                        eclipse_params['primary_width'] = float(eclipse_params['secondary_width'])
                        eclipse_params['primary_depth'] = float(eclipse_params['secondary_depth'])
                        eclipse_params['eclipse_edges'][0] = float(eclipse_params['eclipse_edges'][2])
                        eclipse_params['eclipse_edges'][1] = float(eclipse_params['eclipse_edges'][3])
                        eclipse_params['secondary_position'] = np.nan 
                        eclipse_params['secondary_width'] = np.nan 
                        eclipse_params['secondary_depth'] = np.nan
                        eclipse_params['eclipse_edges'][2] = np.nan
                        eclipse_params['eclipse_edges'][3] = np.nan

        else:
            raise ValueError('Eclipse parameters not computed! Call self.compute_eclipse_params() first.')

        return eclipse_params




from setuptools import setup, find_packages

setup(name='ligeor',
      version='1.0b',
      description='LIght curve GEOmetRy of eclipsing binaries',
      url='https://github.com/gecheline/ligeor',
      author='Angela Kochoska',
      author_email='a.kochoska@gmail.com',
      license='MIT License',
      packages=find_packages(),
      zip_safe=False)


import numpy as np
from ligeor.fitting.sampler import EmceeSampler
from ligeor.models.twogaussian import TwoGaussianModel
from ligeor.utils.lcutils import *


class EmceeSamplerTwoGaussian(EmceeSampler):
    
    def __init__(self, filename='', times = [], fluxes = [], sigmas = [],
                period_init=1, t0_init=0, n_downsample=0, **kwargs):
        '''
        Initializes a sampler for the light curve stored in 'filename'
        with determined initial values for the period and t0.

        Parameters
        ----------
        filename: str
            Filename to load the raw light curve from (expected format: times, fluxes, sigmas)
        times: array-like
            Time points of the observations
        fluxes: array-like
            Observed fluxes
        sigmas: array-like
            Uncertainities of the observed fluxes
        period_init: float
            Initial value of the period (from code/triage)
        t0_init: float
            Initial value of the time of superior conjunction (t0)
        n_downsample: int
            Number of data points to skip in raw light curve for downsampling
        nbins: int
            Number of bins (applies to the computed final model).
        
        Keyword arguments are passed on to load_lc.
            
        
        Attributes
        ----------
        _period_init: float
            Initial value of the orbital period
        _t0_init: float
            Initial value of the time of superior conjunction
        '''
        
        super(EmceeSamplerTwoGaussian,self).__init__(filename, times, fluxes, sigmas, period_init, t0_init, n_downsample, **kwargs)


    def initial_fit(self):
        '''
        Runs an initial fit to the data with the chosen model (two-Gaussian or polyfit).
        
        Attributes
        ----------
        _func: str
            Name of the best fit model function.
        _model_params: array-like
            Labels of all model parameters
        _initial_fit: array-like
            Initial values of the model parameters from optimization.
        '''
        phases, fluxes_ph, sigmas_ph = phase_fold(self._times, 
                                                        self._fluxes, 
                                                        self._sigmas, 
                                                        period=self._period_init, 
                                                        t0=self._t0_init,
                                                        interval = '05')

        twogModel = TwoGaussianModel(phases=phases, 
                                    fluxes=fluxes_ph, 
                                    sigmas=sigmas_ph, 
                                    )
        twogModel.fit()

        # self._twogModel_init = twogModel
        self._func = twogModel.best_fit['func']
        self._model_params = twogModel.best_fit['param_names']
        self._initial_fit = twogModel.best_fit['param_vals']


    def logprob(self, values):
        '''
        Computes the logprobability of the sample.

        Parameters
        ----------
        values: array-like
            period (for phase folding) + model values
        '''

        fmax = self._fluxes.max()
        fmin = self._fluxes.min()
        fdiff = fmax - fmin

        bounds = {'C': ((0),(fmax)),
            'CE': ((0, 1e-6, -0.5),(fmax, fdiff, 0.5)),
            'CG': ((0., -0.5, 0., 0.), (fmax, 0.5, fdiff, 0.5)),
            'CGE': ((0., -0.5, 0., 0., 1e-6, -0.5),(fmax, 0.5, fdiff, 0.5, fdiff, 0.5)),
            'CG12': ((0.,-0.5, 0., 0., -0.5, 0., 0.),(fmax, 0.5, fdiff, 0.5, 0.5, fdiff, 0.5)),
            'CG12E1': ((0.,-0.5, 0., 0., -0.5, 0., 0., 1e-6),(fmax, 0.5, fdiff, 0.5, 0.5, fdiff, 0.5, fdiff)),
            'CG12E2': ((0.,-0.5, 0., 0., -0.5, 0., 0., 1e-6),(fmax, 0.5, fdiff, 0.5, 0.5, fdiff, 0.5, fdiff))}
        
        period, *model_vals = values

        for i,param_val in enumerate(model_vals):
            if param_val < bounds[self._func][0][i] or param_val > bounds[self._func][1][i]:
                # raise Warning('out of prior', self._func, bounds[self._func][0][i], bounds[self._func][1][i], param_val)
                return -np.inf, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan
            
        # fold with period
        phases, fluxes_ph, sigmas_ph = phase_fold(self._times, 
                                                        self._fluxes, 
                                                        self._sigmas, 
                                                        period=period, 
                                                        t0=self._t0_init,
                                                        interval='05')
        # compute model with the selected twog function
        # TODO: review this part for potentially more efficient option
        twog = TwoGaussianModel(phases=phases, 
                                    fluxes=fluxes_ph, 
                                    sigmas=sigmas_ph)
        twog.fit(fit_funcs=[self._func], param_vals=[model_vals])
        
        eclipse_dir = twog.compute_eclipse_params()
        pos1, pos2 = eclipse_dir['primary_position'], eclipse_dir['secondary_position']
        width1, width2 = eclipse_dir['primary_width'], eclipse_dir['secondary_width']
        depth1, depth2 = eclipse_dir['primary_depth'], eclipse_dir['secondary_depth']
        ecl1_area, ecl2_area = twog.eclipse_area[1], twog.eclipse_area[2]
        residuals_mean, residuals_stdev = compute_residuals_stdev(fluxes_ph, twog.model)

        logprob = -0.5*(np.sum((fluxes_ph-twog.model)**2/sigmas_ph**2))
        # print(logprob, pos1, width1, depth1, pos2, width2, depth2)#, ecl1_area, ecl2_area, residuals_mean, residuals_stdev)
        return logprob, pos1, width1, depth1, pos2, width2, depth2, ecl1_area, ecl2_area, residuals_mean, residuals_stdev

    
    def compute_model(self, means, sigmas_low, sigmas_high, save_lc = True, save_file='', show=False, failed=False):
        '''
        Computes the model parameter values from the sample.

        Parameters
        ----------
        means: array-like
            Mean values from the sample
        sigmas_low: array-like
            Standard deviation of samples < mean
        sigmas_high: array_like
            Standard deviation of samples > mean
        save_lc: bool
            If True, saves the model light curve to a file
        save_file: str
            File name to save light curve to, if save_lc=True.
        show: bool
            If True, will display a plot of the model light curve.
        failed: bool
            If True, all model parameters are np.nan
        '''        
        
        model_results = {'C': np.nan, 'mu1': np.nan, 'd1': np.nan, 'sigma1': np.nan, 
                        'mu2': np.nan, 'd2': np.nan, 'sigma2': np.nan, 'Aell': np.nan, 'phi0': np.nan
                        }
        model_results_err = {'C': np.nan, 'mu1': np.nan, 'd1': np.nan, 'sigma1': np.nan, 
                        'mu2': np.nan, 'd2': np.nan, 'sigma2': np.nan, 'Aell': np.nan, 'phi0': np.nan
                        }
        
        # results_str = '{}'.format(func)
        if failed:
            self._model_values = model_results
            self._model_values_errs = model_results_err
            self._chi2 = np.nan
        
        else:
            for mkey in model_results.keys():
                if mkey in self._model_params:
                    pind = self._model_params.index(mkey)
                    model_results[mkey] = means[pind+1]
                    model_results_err[mkey] = np.max((sigmas_low[pind+1],sigmas_high[pind+1]))
                    # results_str += ',{},{}'.format(model_results[mkey],model_results_err[mkey])
                
            self._model_values = model_results
            self._model_values_errs = model_results_err
            chi2 = np.nan

            phases_obs, fluxes_ph_obs, sigmas_ph_obs = phase_fold(self._times, 
                                                    self._fluxes, 
                                                    self._sigmas, 
                                                    period=self._period_mcmc['value'], 
                                                    t0=self._t0_init, interval='05')
            twog_func = getattr(TwoGaussianModel, self._func.lower())
            fluxes_model = twog_func(phases_obs, *means[1:])
            chi2 = -0.5*(np.sum((fluxes_ph_obs-fluxes_model)**2/sigmas_ph_obs**2))
            
            if show:
                import matplotlib.pyplot as plt
                plt.plot(phases_obs, fluxes_ph_obs, 'k.')
                plt.plot(phases_obs, fluxes_model, 'r-')
                plt.show()

            if save_lc:
                np.savetxt(save_file, np.array([phases_obs, fluxes_model]).T)

            self._chi2 = chi2